# Handles coordinate systems
# 
# Copyright (c) 2022, herculens developers and contributors

__author__ = 'aymgal'


import os
import shutil
import math
import numpy as np
from astropy.io import fits

from herculens.Inference.legacy.parameters import Parameters as HerculensParameters
from herculens.Util import param_util
from herculens.Util.jax_util import unjaxify_kwargs

from coolest.template.lazy import *


# NOTE: `h2c` is a shorthand for `herculens2coolest`


def create_directory(directory, empty_dir_bool):
    if not os.path.exists(directory):
        os.makedirs(directory)
        print(f"COOLEST-info: Created directory {directory}")
    elif empty_dir_bool:
        shutil.rmtree(directory)
        os.makedirs(directory)
        print(f"COOLEST-info: Emptied directory {directory}")
    elif not os.listdir(directory):
        print(f"COOLEST-info: Directory {directory} already empty; nothing to be done.")
    else:
        return False
    return True


def skycoord_to_coolest(skycoord):
    ra  = skycoord.ra.to_string()
    dec = skycoord.dec.to_string()
    return CoordinatesOrigin(ra=ra, dec=dec)


def create_observation(data, lens_image, json_dir=None,
                       noise_type='NoiseMap', model_noise_map=None, 
                       kwargs_obs=None, kwargs_noise=None):
    # saves the observed lens image on disk
    obs_file_name = "obs.fits"
    if json_dir is not None:
        obs_fits_path = os.path.join(json_dir, obs_file_name)
    else:
        obs_fits_path = obs_file_name
    save_image_to_fits(obs_fits_path, data, 
                       header_cards=[('COOLEST', "Observed lens image")])
    model_grid = lens_image.Grid
    if model_grid.x_is_inverted or model_grid.y_is_inverted:
        raise NotImplementedError("Grid orientation not yet supported")
    extent = model_grid.extent
    pix_scl = model_grid.pixel_width
    fov_x = [extent[0] - pix_scl/2., extent[1] + pix_scl/2.]
    fov_y = [extent[2] - pix_scl/2., extent[3] + pix_scl/2.]
    pixels = PixelatedRegularGrid(obs_file_name,
                                  field_of_view_x=fov_x,
                                  field_of_view_y=fov_y,
                                  fits_file_dir=json_dir)

    if kwargs_noise is None:
        kwargs_noise = {}
    if noise_type == 'NoiseMap':
        if model_noise_map is None:
            raise ValueError(f"A noise map must be provided for noise type {noise_type}")
        noise_map_file_name = "noise_map.fits"
        if json_dir is not None:
            noise_map_fits_path = os.path.join(json_dir, noise_map_file_name)
        else:
            noise_map_fits_path = noise_map_file_name
        save_image_to_fits(noise_map_fits_path, model_noise_map, 
                           header_cards=[('COOLEST', "Model noise map")])
        noise_map = PixelatedRegularGrid(noise_map_file_name,
                                         field_of_view_x=fov_x,
                                         field_of_view_y=fov_y,
                                         fits_file_dir=json_dir)
        noise = NoiseMap(noise_map, **kwargs_noise)
        assert math.isclose(pixels.pixel_size, noise_map.pixel_size,
                            rel_tol=1e-09, abs_tol=0.0)
    else:
        raise NotImplementedError(f"Noise type {noise_type} not yet supported")
    
    exp_map = lens_image.Noise.exposure_map
    if isinstance(exp_map, (int, float)):
        pass
    elif isinstance(exp_map, np.ndarray) and exp_map.ndim == 2:
        exp_map_file_name = "exposure_map"
        exp_map = PixelatedRegularGrid(exp_map_file_name,
                                        field_of_view_x=fov_x,
                                        field_of_view_y=fov_y,
                                        fits_file_dir=json_dir)
    else:
        raise NotImplementedError("Exposure time / map can only be a float or integer, or a 2D array")
    
    if kwargs_obs is None:
        kwargs_obs = {}
    observation = Observation(pixels=pixels,
                              exposure_time=exp_map,
                              noise=noise,
                              **kwargs_obs)
    return observation


def create_instrument(lens_image, observation, json_dir=None,
                      psf_type='PixelatedPSF',
                      psf_description=None,
                      kwargs=None):
    if psf_type == 'PixelatedPSF':
        super_factor = lens_image.PSF.kernel_supersampling_factor
        if super_factor > 1:
            psf_kernel = lens_image.PSF.kernel_point_source_supersampled(super_factor)
        elif super_factor == 1:
            psf_kernel = lens_image.PSF.kernel_point_source
        else:
            raise ValueError(f"PSF supersampling factor of {super_factor} is unknown "
                             f"for `psf_type` {psf_type}")
        psf_file_name = "psf.fits"
        if json_dir is not None:
            psf_fits_path = os.path.join(json_dir, psf_file_name)
        else:
            psf_fits_path = psf_file_name
        save_image_to_fits(psf_fits_path, psf_kernel, 
                           header_cards=[('COOLEST', "Pixelated PSF kernel")])
        pixel_size = lens_image.Grid.pixel_width / super_factor
        nx, ny = psf_kernel.shape
        fov_x = (0, nx * pixel_size)
        fov_y = (0, ny * pixel_size)
        pixels = PixelatedRegularGrid(psf_file_name,
                                      field_of_view_x=fov_x,
                                      field_of_view_y=fov_y,
                                      fits_file_dir=json_dir)
        if psf_description is None:
            psf_description = ""
        psf = PixelatedPSF(pixels, description=psf_description)
    else:
        raise NotImplementedError(f"Noise type {noise_type} not yet supported")

    if kwargs is None:
        kwargs = {}
    instrument = Instrument(pixel_size,
                            psf=psf,
                            **kwargs)

    assert math.isclose(observation.pixels.pixel_size, instrument.pixel_size,
                        rel_tol=1e-09, abs_tol=0.0)

    return instrument


def create_likelihoods(lens_image, json_dir=None, 
                       likelihood_type='ImagingDataLikelihood', 
                       likelihood_mask=None):
    likelihood_list = []
    if likelihood_type == 'ImagingDataLikelihood':
        if likelihood_mask is not None:
            # saves the observed lens image on disk
            obs_file_name = "mask-likelihood.fits"
            if json_dir is not None:
                obs_fits_path = os.path.join(json_dir, obs_file_name)
            else:
                obs_fits_path = obs_file_name
            save_image_to_fits(obs_fits_path, likelihood_mask, 
                            header_cards=[('COOLEST', "Likelihood mask (0 are excluded pixels)")])
            model_grid = lens_image.Grid
            if model_grid.x_is_inverted or model_grid.y_is_inverted:
                raise NotImplementedError("Grid orientation not yet supported")
            extent = model_grid.extent
            pix_scl = model_grid.pixel_width
            fov_x = [extent[0] - pix_scl/2., extent[1] + pix_scl/2.]
            fov_y = [extent[2] - pix_scl/2., extent[3] + pix_scl/2.]
            mask = PixelatedRegularGrid(obs_file_name,
                                        field_of_view_x=fov_x,
                                        field_of_view_y=fov_y,
                                        fits_file_dir=json_dir)
            likelihood_list.append(ImagingDataLikelihood(mask=mask))
    else:
        raise NotImplementedError(f"Likelihood type {likelihood_type} not yet supported") 
    return DataLikelihoodList(*likelihood_list)


def save_image_to_fits(path, image, header_cards=[], overwrite=True):
    header = fits.Header(cards=header_cards)
    fits.writeto(path, image, header, overwrite=overwrite)
    print(f"COOLEST-info: Saved image to FITS file {path}")


def update_lensing_entities(lens_image, lensing_entity_mapping, 
                            parameters=None, samples=None, 
                            current_entities=None, re_create_entities=True,
                            json_dir=None, fits_file_suffix="coolest"):
    """
    lensing_entity_mapping: list of 2-tuples of the following format:
        ('name_of_the_entity', kwargs_mapping)
    where kwargs_mapping is settings for update_mass_field_model and update_galaxy_model functions. 
    """
    # TODO: check if multi-plane lensing

    if re_create_entities is False:
        if current_entities is None or len(current_entities) == 0:
            print("COOLEST-info: Since no lensing entities have been found nor provided, "
                         "new ones will be created based on the model.")
            re_create_entities = True
        else:
            print("COOLEST-info: Current lensing entities will be updated.")
    else:
        print("COOLEST-info: New lensing entities will be created based on the model.")

    if parameters is not None and isinstance(parameters, dict):
        parameters = unjaxify_kwargs(parameters)
    if samples is not None and isinstance(samples, dict):
        samples = unjaxify_kwargs(samples)

    # initialize list of lensing entities
    entities = []

    # iterate over the lensing entities (galaxies or external shears)
    for i, (entity_name, kwargs_mapping) in enumerate(lensing_entity_mapping):
        if kwargs_mapping is None:
            continue  # this entity is ignored (neither created nor updated!)
        
        entity_type = kwargs_mapping.pop('type')

        if re_create_entities is True:
            current_entity = None
        else:
            current_entity = current_entities[i]
        
        if entity_type == 'MassField':
            entity = update_mass_field_model(lens_image, entity_name,
                                             mass_field=current_entity, 
                                             parameters=parameters,
                                             samples=samples,
                                             **kwargs_mapping)
        elif entity_type == 'Galaxy':
            entity = update_galaxy_model(lens_image, entity_name, 
                                         galaxy=current_entity, 
                                         parameters=parameters,
                                         samples=samples,
                                         file_dir=json_dir,
                                         fits_file_suffix=fits_file_suffix,
                                         **kwargs_mapping)
        else:
            raise ValueError(f"Unknown lensing entity type '{entity_type}'.")

        if re_create_entities:
            entities.append(entity)
        else:
            # in that case the current entity has been updated in place
            pass

    return current_entities if not re_create_entities else LensingEntityList(*entities)


def update_mass_field_model(lens_image, name, mass_field=None,
                            parameters=None, samples=None,
                            mass_profile_indices=None, 
                            redshift=None):
    # external shear
    mass_profiles_all = lens_image.MassModel.profile_type_list
    mass_profiles_in  = [mass_profiles_all[i] for i in mass_profile_indices]
    if mass_profiles_in not in (['SHEAR'], ['SHEAR_GAMMA_PSI']):
        raise NotImplementedError("Mass fields other than a single external shear are not yet supported.")
    mass_profiles_out = h2c_extshear_profiles(mass_profiles_in)
    
    # instantiate the external shear
    if mass_field is None:
        mass_field = MassField(name, mass_model=MassModel(*mass_profiles_out), redshift=redshift)
    else:
        assert isinstance(mass_field, MassField), "Inconsistent entity type"

    kwargs_all, kwargs_all_samples = get_parameters_and_samples(parameters, samples)
        
    if kwargs_all is not None or kwargs_all_samples is not None:
        # add point estimate values to the MassField object
        for ic, ih in enumerate(mass_profile_indices):
            profile_name = mass_profiles_in[ic]
            if profile_name == 'SHEAR':
                g1g2_param = True
            elif profile_name == 'SHEAR_GAMMA_PSI':
                g1g2_param = False
            if kwargs_all is not None:
                h2c_extshear_values(mass_field.mass_model[ic], 
                                    kwargs_all['kwargs_lens'][ih], 
                                    g1g2_param=g1g2_param)
            if kwargs_all_samples is not None:
                h2c_extshear_posteriors(mass_field.mass_model[ic], 
                                        kwargs_all_samples['kwargs_lens'][ih], 
                                        g1g2_param=g1g2_param)

    return mass_field


def update_galaxy_model(lens_image, name, galaxy=None,
                        parameters=None, samples=None,
                        mass_profile_indices=None, light_profile_indices=None,
                        lensed=None, redshift=None, 
                        file_dir=None, fits_file_suffix="coolest"):
    # mass model
    if mass_profile_indices is not None:
        mass_profiles_all = lens_image.MassModel.profile_type_list
        mass_profiles_in  = [mass_profiles_all[i] for i in mass_profile_indices]
        mass_profiles_out = h2c_mass_profiles(mass_profiles_in)
    else:
        mass_profiles_out = []

    # light model
    if light_profile_indices is not None:
        if lensed:
            light_profiles_all = lens_image.SourceModel.profile_type_list
        else:
            light_profiles_all = lens_image.LensLightModel.profile_type_list
        light_profiles_in  = [light_profiles_all[i] for i in light_profile_indices]
        light_profiles_out = h2c_light_profiles(light_profiles_in)
    else:
        light_profiles_out = []
    
    if galaxy is None:
        # instantiate the galaxy
        galaxy = Galaxy(name,
                        mass_model=MassModel(*mass_profiles_out), 
                        light_model=LightModel(*light_profiles_out),
                        redshift=redshift)
    else:
        assert isinstance(galaxy, Galaxy), "Inconsistent entity type"

    kwargs_all, kwargs_all_samples = get_parameters_and_samples(parameters, samples)

    if kwargs_all is not None or kwargs_all_samples is not None:
        if mass_profile_indices is not None:
            update_galaxy_mass_model(galaxy, lens_image, kwargs_all, kwargs_all_samples,
                                     mass_profile_indices, mass_profiles_in, 
                                     file_dir=file_dir, fits_file_suffix=fits_file_suffix)
        if light_profile_indices is not None:
            update_galaxy_light_model(galaxy, lens_image, kwargs_all, kwargs_all_samples,
                                      light_profile_indices, light_profiles_in, lensed,
                                      file_dir=file_dir, fits_file_suffix=fits_file_suffix)

    return galaxy


def update_galaxy_mass_model(galaxy, lens_image, kwargs_all, kwargs_all_samples, 
                             profile_indices, profile_names, 
                             file_dir=None, fits_file_suffix="coolest"):
    comp_name = 'lens_mass'
    key = 'kwargs_lens'
    # add point estimate values
    for ic, ih in enumerate(profile_indices):
        profile_name = profile_names[ic]
        if profile_name in ['SIS', 'SIE', 'EPL', 'PEMD', 'SPEMD']:
            if profile_name == 'SIS':
                isothermal, spherical = True, True
            elif profile_name == 'SIE':
                isothermal, spherical = True, False
            elif profile_name in ['EPL', 'PEMD']:
                isothermal, spherical = False, False
            else:
                raise NotImplementedError(f"Parametric mass profile '{profile_name}' is not yet supported")
            if kwargs_all is not None:
                h2c_powerlaw_values(galaxy.mass_model[ic], kwargs_all[key][ih], 
                                    isothermal=isothermal, spherical=spherical)
            if kwargs_all_samples.get(key, None) is not None:
                h2c_powerlaw_posteriors(galaxy.mass_model[ic], kwargs_all_samples[key][ih], 
                                        isothermal=isothermal, spherical=spherical)

        elif profile_name in ['PIXELATED', 'PIXELATED_DIRAC', 'PIXELATED_FIXED']:
            galaxy_name = galaxy.name.replace(' ', '_')
            if profile_name == 'PIXELATED_FIXED':
                if kwargs_all is not None:
                    h2c_fixed_pixelated_lens_mass(
                        galaxy.mass_model[ic],   # coolest profile
                        lens_image.MassModel.func_list[ih],  # herculens profile
                        kwargs_all[key][ih],
                        file_dir=file_dir,
                        fits_file_suffix=fits_file_suffix+f'-{galaxy_name}_{ic}'
                    )
                if kwargs_all_samples.get(key, None) is not None:
                    print(f"COOLEST-warning: Posterior stats for profile '{profile_names[ic]}' "
                        f"is not yet supported; samples are ignored.")
            else:
                raise NotImplementedError(f"Pixelated mass profile '{profile_name}' is not yet supported")
        
        else:
            raise NotImplementedError(f"'{profile_name}' not yet supported.")


def h2c_fixed_pixelated_lens_mass(coolest_profile, herculens_profile, 
                                  kwargs_lens_profile, file_dir=None, 
                                  fits_file_suffix="coolest"):
    for herculens_key, coolest_key, stacked in (
            ('func_pixels' , 'pixels', False), 
            ('deriv_pixels', 'pixels_derivative', True), 
            ('hess_pixels' , 'pixels_hessian', True),
        ):
        x_grid, y_grid = herculens_profile.get_pixel_grid(component=herculens_key).pixel_coordinates
        if stacked is True:
            # 3D array, with x and y axes corresponding to axis 1 and 2
            pixels = np.stack(kwargs_lens_profile[herculens_key], axis=0)
        else:
            # 2D array, with x and y axes corresponding to axis 0 and 1
            pixels = kwargs_lens_profile[herculens_key]
        h2c_pixelated_values(coolest_profile, 
                             pixels, 
                             x_grid, y_grid,
                             param_key=coolest_key,
                             file_dir=file_dir,
                             fits_file_suffix=fits_file_suffix+f'_{coolest_key}',
                             description="Pixelated lens potential and derivatives",
                             )  # TODO: improve access to e.g. pixel_grid 


def update_galaxy_light_model(galaxy, lens_image, kwargs_all, kwargs_all_samples, 
                              profile_indices, profile_names, lensed, 
                              file_dir=None, fits_file_suffix="coolest"):
    # get current values
    class_comp = lens_image.SourceModel if lensed else lens_image.LensLightModel
    comp_name = 'source' if lensed else 'lens_light'
    key = 'kwargs_source' if lensed else 'kwargs_lens_light'
    # add point estimate values
    for ic, ih in enumerate(profile_indices):
        if profile_names[ic] == 'SERSIC_ELLIPSE':
            if kwargs_all is not None:
                h2c_Sersic_values(galaxy.light_model[ic], kwargs_all[key][ih])
            if kwargs_all_samples.get(key, None) is not None:
                h2c_Sersic_posteriors(galaxy.light_model[ic], kwargs_all_samples[key][ih])
        elif profile_names[ic] == 'SERSIC':
            if kwargs_all is not None:
                h2c_Sersic_values(galaxy.light_model[ic], kwargs_all[key][ih], spherical=True)
            if kwargs_all_samples.get(key, None) is not None:
                h2c_Sersic_posteriors(galaxy.light_model[ic], kwargs_all_samples[key][ih], spherical=True)
        elif profile_names[ic] == 'SHAPELETS':
            if kwargs_all is not None:
                h2c_Shapelets_values(galaxy.light_model[ic], kwargs_all[key][ih],
                                     class_comp.func_list[ih])  # TODO: improve access to e.g. n_max 
            if kwargs_all_samples.get(key, None) is not None:
                print(f"COOLEST-warning: Posterior stats for profile '{profile_names[ic]}' "
                      f"is not yet supported; samples are ignored.")
        elif profile_names[ic] == 'UNIFORM':
            if kwargs_all is not None:
                h2c_Uniform_values(galaxy.light_model[ic], kwargs_all[key][ih])
            if kwargs_all_samples.get(key, None) is not None:
                h2c_Uniform_posteriors(galaxy.light_model[ic], kwargs_all_samples[key][ih])
        elif profile_names[ic] == 'PIXELATED':
            if kwargs_all is not None:
                x_grid_model, y_grid_model, _ = lens_image.get_source_coordinates(kwargs_all['kwargs_lens'])
                h2c_pixelated_values(galaxy.light_model[ic], 
                                     kwargs_all[key][ih]['pixels'],
                                     x_grid_model, y_grid_model,
                                     file_dir=file_dir,
                                     fits_file_suffix=fits_file_suffix+f'-{comp_name}_{ic}',
                                     description="Pixelated surface brightness model",
                                     )  # TODO: improve access to e.g. pixel_grid 
            if kwargs_all_samples.get(key, None) is not None:
                print(f"COOLEST-warning: Posterior stats for profile '{profile_names[ic]}' "
                      f"is not yet supported; samples are ignored.")
        else:
            raise NotImplementedError(f"'{profile_names[ic]}' not yet supported.")


def h2c_powerlaw_values(profile, kwargs, isothermal=False, spherical=False):
    theta_E  = check_type(kwargs['theta_E'])
    center_x = check_type(kwargs['center_x'])
    center_y = check_type(kwargs['center_y'])
    if spherical:
        phi, q = 0., 1.
    else:
        e1 = check_type(kwargs['e1'])
        e2 = check_type(kwargs['e2'])
        phi, q = param_util.ellipticity2phi_q(e1, e2)
        phi = h2c_position_angle(phi)
        phi = check_type(phi)
        q = check_type(q)
    if not isothermal:
        gamma = check_type(kwargs['gamma'])
    profile.parameters['theta_E'].set_point_estimate(theta_E)
    profile.parameters['center_x'].set_point_estimate(center_x)
    profile.parameters['center_y'].set_point_estimate(center_y)
    profile.parameters['phi'].set_point_estimate(phi)
    profile.parameters['q'].set_point_estimate(q)
    if spherical:
        profile.parameters['phi'].fix()
        profile.parameters['q'].fix()
    if not isothermal:
        profile.parameters['gamma'].set_point_estimate(gamma)


def h2c_powerlaw_posteriors(profile, kwargs_samples, isothermal=False, spherical=False):
    profile.parameters['theta_E'].set_posterior(prepare_posterior(kwargs_samples['theta_E']))
    profile.parameters['center_x'].set_posterior(prepare_posterior(kwargs_samples['center_x']))
    profile.parameters['center_y'].set_posterior(prepare_posterior(kwargs_samples['center_y']))
    if spherical:
        phi, q = 0., 1.
    else:
        e1 = kwargs_samples['e1']
        e2 = kwargs_samples['e2']
        if e1 is not None and e2 is not None:
            phi, q = param_util.ellipticity2phi_q_numpy(e1, e2)
            phi = h2c_position_angle(phi)
        else:
            phi, q = None, None
    profile.parameters['phi'].set_posterior(prepare_posterior(phi))
    profile.parameters['q'].set_posterior(prepare_posterior(q))
    if not isothermal:
        profile.parameters['gamma'].set_posterior(prepare_posterior(kwargs_samples['gamma']))


def h2c_Sersic_values(profile, kwargs, spherical=False):
    amp = check_type(kwargs['amp'])
    R_sersic = check_type(kwargs['R_sersic'])
    n_sersic = check_type(kwargs['n_sersic'])
    center_x = check_type(kwargs['center_x'])
    center_y = check_type(kwargs['center_y'])
    if spherical:
        phi, q = 0., 1.  # spherical case
    else:
        e1 = check_type(kwargs['e1'])
        e2 = check_type(kwargs['e2'])
        phi, q = param_util.ellipticity2phi_q(e1, e2)
        phi = h2c_position_angle(phi)
        phi = check_type(phi)
        q = check_type(q)
    R_sersic = convert_major_axis_radius(R_sersic, q)
    profile.parameters['I_eff'].set_point_estimate(amp)
    profile.parameters['theta_eff'].set_point_estimate(R_sersic)
    profile.parameters['n'].set_point_estimate(n_sersic)
    profile.parameters['center_x'].set_point_estimate(center_x)
    profile.parameters['center_y'].set_point_estimate(center_y)
    profile.parameters['phi'].set_point_estimate(phi)
    profile.parameters['q'].set_point_estimate(q)
    if spherical:
        profile.parameters['phi'].fix()
        profile.parameters['q'].fix()


def h2c_Sersic_posteriors(profile, kwargs_samples, spherical=False):
    profile.parameters['I_eff'].set_posterior(prepare_posterior(kwargs_samples['amp']))
    profile.parameters['n'].set_posterior(prepare_posterior(kwargs_samples['n_sersic']))
    profile.parameters['center_x'].set_posterior(prepare_posterior(kwargs_samples['center_x']))
    profile.parameters['center_y'].set_posterior(prepare_posterior(kwargs_samples['center_y']))
    if spherical:
        phi, q = 0., 1.  # spherical case
    else:
        e1 = kwargs_samples['e1']
        e2 = kwargs_samples['e2']
        if e1 is not None and e2 is not None:
            phi, q = param_util.ellipticity2phi_q_numpy(e1, e2)
            phi = h2c_position_angle(phi)
        else:
            phi, q = None, None
    R_sersic = kwargs_samples['R_sersic']
    R_sersic = convert_major_axis_radius(R_sersic, q)
    profile.parameters['theta_eff'].set_posterior(prepare_posterior(kwargs_samples['R_sersic']))
    profile.parameters['phi'].set_posterior(prepare_posterior(phi))
    profile.parameters['q'].set_posterior(prepare_posterior(q))


def h2c_Uniform_values(profile, kwargs):
    amp = check_type(kwargs['amp'])
    profile.parameters['A'].set_point_estimate(amp)


def h2c_Uniform_posteriors(profile, kwargs_samples):
    profile.parameters['A'].set_posterior(prepare_posterior(kwargs_samples['amp']))


def h2c_Shapelets_values(profile, kwargs, profile_herculens):
    amps = check_type(kwargs['amps'])
    beta = check_type(kwargs['beta'])
    center_x = check_type(kwargs['center_x'])
    center_y = check_type(kwargs['center_y'])
    n_max = check_type(profile_herculens.maximum_order)
    profile.parameters['amps'].set_point_estimate(amps)
    profile.parameters['beta'].set_point_estimate(beta)
    profile.parameters['center_x'].set_point_estimate(center_x)
    profile.parameters['center_y'].set_point_estimate(center_y)
    profile.parameters['n_max'].set_point_estimate(int(n_max))
    profile.parameters['n_max'].fix()


def h2c_pixelated_values(profile, pixel_values_in, x_grid, y_grid, param_key='pixels',
                         file_dir=None, fits_file_suffix="coolest", 
                         description="Profile defined on a pixelated grid"):
    """
    Profile based on REGULAR grid of pixels.
    
    `pixels` can be either a regular 2D arrays of pixels, or a stack (3D array) of
    pixel grids defined over the same coordinates grid. 
    """
    pixel_values = check_type(pixel_values_in)
    if x_grid[0, 0] > x_grid[0, 1]:
        raise NotImplementedError("Only increasing x coordinates is supported so far")
    if y_grid[0, 0] > y_grid[1, 0]:
        raise NotImplementedError("Only increasing y coordinates is supported so far")
    pixel_scale = float(abs(x_grid[0, 0] - x_grid[0, 1]))
    pixel_scale_y = float(abs(y_grid[0, 0] - y_grid[1, 0]))
    print(f"COOLEST-info: assuming square pixels for the pixelated grid "
          f"(pixel size along x = {pixel_scale}, along y = {pixel_scale_y}).")
    
    half_pix = pixel_scale / 2.
    extent = [x_grid[0, 0], x_grid[0, -1], y_grid[0, 0], y_grid[-1, 0]]
    fov_x = [extent[0]-half_pix, extent[1]+half_pix]
    fov_y = [extent[2]-half_pix, extent[3]+half_pix]

    transform_pix2angle = np.array([[pixel_scale, 0.], [0, pixel_scale]])
    matrix = transform_pix2angle / 3600.  # arcsec -> degree
    CD1_1 = float(matrix[0, 0])
    CD1_2 = float(matrix[0, 1])
    CD2_1 = float(matrix[1, 0])
    CD2_2 = float(matrix[1, 1])

    fits_filename = f'model-{fits_file_suffix}.fits'
    if file_dir is None:
        fits_path = fits_filename
    else:
        fits_path = os.path.join(file_dir, fits_filename)
    save_image_to_fits(fits_path, pixel_values, 
                       header_cards=[
                            ('COOLEST', description),
                            ('PIXSCALE', pixel_scale),
                            ('CD1_1', CD1_1),
                            ('CD1_2', CD1_2),
                            ('CD2_1', CD2_1),
                            ('CD2_2', CD2_2),
                       ])
    if len(pixel_values.shape) == 3:
        pixels_parameters = PixelatedRegularGridStack(
            fits_filename, # relative path to fits file
            field_of_view_x=fov_x,
            field_of_view_y=fov_y,
            check_fits_file=True,
            fits_file_dir=file_dir,
        )
    else: 
        pixels_parameters = PixelatedRegularGrid(
            fits_filename, # relative path to fits file
            field_of_view_x=fov_x,
            field_of_view_y=fov_y,
            check_fits_file=True,
            fits_file_dir=file_dir,
        )
    # set the pixel values
    profile.parameters[param_key] = pixels_parameters
    # finally, just perform a consistency check regarding the pixel size
    recomputed_pixel_scale = profile.parameters[param_key].pixel_size
    consistent = math.isclose(pixel_scale, recomputed_pixel_scale, 
                              rel_tol=0.0, abs_tol=1e-06)
    if not consistent:
        raise ValueError(f"Inconsistent pixel scale values for profile '{profile.type}' "
                         f"({pixel_scale} vs. {recomputed_pixel_scale})")


def h2c_extshear_values(profile, kwargs, g1g2_param=False):
    if g1g2_param:
        gamma1, gamma2 = kwargs['gamma1'], kwargs['gamma2']
        phi_ext, gamma_ext = param_util.shear_cartesian2polar(gamma1, gamma2)
    else:
        phi_ext, gamma_ext = kwargs['phi_ext'], kwargs['gamma_ext']
    phi_ext   = check_type(phi_ext)
    gamma_ext = check_type(gamma_ext)
    phi_ext = h2c_position_angle(phi_ext)
    profile.parameters['phi_ext'].set_point_estimate(phi_ext)
    profile.parameters['gamma_ext'].set_point_estimate(gamma_ext)


def h2c_extshear_posteriors(profile, kwargs_samples, g1g2_param=False):
    if g1g2_param:
        gamma1, gamma2 = kwargs_samples['gamma1'], kwargs_samples['gamma2']
        phi_ext, gamma_ext = param_util.shear_cartesian2polar_numpy(gamma1, gamma2)
    else:
        phi_ext, gamma_ext = kwargs_samples['phi_ext'], kwargs_samples['gamma_ext']
    phi_ext = h2c_position_angle(phi_ext)
    profile.parameters['phi_ext'].set_posterior(prepare_posterior(phi_ext))
    profile.parameters['gamma_ext'].set_posterior(prepare_posterior(gamma_ext))


def h2c_extshear_profiles(profiles_herculens):
    profiles_coolest = []
    for i, profile_herculens in enumerate(profiles_herculens):
        if profile_herculens in ['SHEAR', 'SHEAR_GAMMA_PSI']:
            profiles_coolest.append('ExternalShear')
        else:
            raise ValueError(f"Unknown COOLEST mapping for mass profile '{profile_herculens}'.")
    return profiles_coolest


def h2c_mass_profiles(profiles_herculens):
    profiles_coolest = []
    for profile_herculens in profiles_herculens:
        if profile_herculens in ['SIS', 'SIE']:
            profiles_coolest.append('SIE')
        elif profile_herculens in ['NIE']:
            profiles_coolest.append('NIE')
        elif profile_herculens in ['PEMD', 'EPL']:
            profiles_coolest.append('PEMD')
        elif profile_herculens in ['SPEMD']:
            profiles_coolest.append('SPEMD')
        elif profile_herculens in ['PIXELATED']:
            profiles_coolest.append('PixelatedRegularGridPotential')
        elif profile_herculens in ['PIXELATED_FIXED']:
            profiles_coolest.append('PixelatedRegularGridFullyDefined')
        elif profile_herculens in ['SHEAR', 'SHEAR_GAMMA_PSI']:
            raise ValueError("External shear cannot be a galaxy mass profile.")
        else:
            raise ValueError(f"Unknown COOLEST mapping for mass profile '{profile_herculens}'.")
    return profiles_coolest


def h2c_light_profiles(profiles_herculens):
    profiles_coolest = []
    for profile_herculens in profiles_herculens:
        if profile_herculens in ['SERSIC', 'SERSIC_ELLIPSE']:
            profiles_coolest.append('Sersic')
        elif profile_herculens in ['UNIFORM']:
            profiles_coolest.append('Uniform')
        elif profile_herculens in ['SHAPELETS']:
            profiles_coolest.append('Shapelets')
        elif profile_herculens in ['SHAPELETS']:
            profiles_coolest.append('LensedPS')
        elif profile_herculens in ['PIXELATED']:
            profiles_coolest.append('PixelatedRegularGrid')
        else:
            raise ValueError(f"Unknown COOLEST mapping for light profile '{profile_herculens}'.")
    return profiles_coolest


def h2c_position_angle(value):
    """
    Transform an angle in radian from Herculens into an angle in degree in the COOLEST conventions.
    Based on @LyneVdV's implementation for lenstronomy.
    """
    value_conv = value + np.pi / 2.
    value_conv = value_conv * 180. / np.pi
    print("COOLEST-warning: Assuming x axis is positive towards the right when converting angles!")
    if is_iterable(value):
       median = np.median(value)
       for i, val in enumerate(value_conv):
            if median <= -90.:
                value_conv[i] += 180.
            elif median > 90.:
               value_conv[i] -= 180.
    else:
        if value_conv <= -90:
            value_conv += 180.
        elif value_conv > 90:
            value_conv -= 180.
    return value_conv


def prepare_posterior(samples):
    if samples is None or isinstance(samples, (int, float)) or len(samples) == 1:
        mean = samples
        perc = None, None, None
    else:
        mean, perc = np.mean(samples), np.percentile(samples, q=[16, 50, 84])
    posterior = PosteriorStatistics(mean=check_type(mean),
                                    median=check_type(perc[1]),
                                    percentile_16th=check_type(perc[0]),
                                    percentile_84th=check_type(perc[2]))
    return posterior


def convert_major_axis_radius(r, q):
    return r * np.sqrt(q)


def check_type(value):
    if value is None:
        return None
    if is_iterable(value):
        value_valid = np.asarray(value)
    else:
        value_valid = float(value)
    return value_valid


def is_iterable(value):
    try:
        _ = iter(value)  # test if value is iterable
    except:
        return False
    else:
        return True

def get_parameters_and_samples(parameters, samples):
    kwargs_all = None
    kwargs_all_samples = None

    if parameters is not None:
        if isinstance(parameters, HerculensParameters):
            print("COOLEST-info: Using the legacy interface of the Parameters class")
            kwargs_all = parameters.best_fit_values(as_kwargs=True)
            if parameters.samples is not None:
                kwargs_all_samples = parameters.samples(as_kwargs=True, group_by_param=True)
            else:
                kwargs_all_samples = None

        elif isinstance(parameters, dict):
            kwargs_all = parameters
            kwargs_all_samples = None if samples is None else samples

    elif parameters is None and samples is not None:
        kwargs_all = None
        if isinstance(samples, dict):
            kwargs_all_samples = None if samples is None else samples
        else:
            raise ValueError(f"Unsupported type '{type(samples)}' for `samples`")
    
    return kwargs_all, kwargs_all_samples



# Defines the model of a multi-plane strong lens
# 
# Copyright (c) 2024, herculens developers and contributors

__author__ = 'krawczyk'

import numpy as np
import jax
import jax.numpy as jnp
import scipy.ndimage

from functools import partial
from herculens.LensImage.Numerics.numerics import Numerics


class MPLensImage(object):
    """Generate a multi-plane lensed images from source light, lens mass/light, and point source models."""

    def __init__(
        self,
        grid_class,
        psf_class,
        noise_class,
        mass_model_class,
        light_model_class,
        source_arc_masks=None,
        source_grid_scale=None,
        conjugate_points=None,
        kwargs_numerics=None
    ):
        '''Generate a multi-plane lensed images from source light, lens mass/light, and point source models.

        Parameters
        ----------
        grid_class : PixelGrid
            coordinate system, instance of PixelGrid() from herculens.Coordinates.pixel_grid
        psf_class : PSF
            point spread function, instance of PSF() from herculens.Instrument.psf
        noise_class : Noise
            noise properties, instance of Noise() from herculens.Instrument.noise
        mass_model_class : MPMassModel
            multi-plane mass model, instance of MPMassModel() from
            herculens.MassModel.mass_model_multiplane
        light_model_class : MPLightModel
            multi-plane light model, instance of MPLightModel() from
            herculens.LightModel.light_model_multiplane
        source_arc_masks : List of array_like, optional
            list of 2D boolean array to define the region over which the
            (pixelated) lensed source is modeled, one for each plane, by default None
        source_grid_scale : float, optional
            a float between 0 and 1 indicating a global scale factor for pixelated grids
            (e.g. 0.5 will set the pixel grid to be 50% the extent defined by the arc mask
            traced back to the source plane), by default None
        conjugate_points : list of lists, optional
            a list of lists of arrays that can be traced back to each plane
            with the `MPLensImage.trace_conjugate_points` method, by default None
        kwargs_numerics : dict, optional
            keyword arguments for various numerical settings (see herculens.Numerics.numerics),
            by default None
        '''
        self.Grid = grid_class
        self.PSF = psf_class
        self.Noise = noise_class
        self.PSF.set_pixel_size(self.Grid.pixel_width)
        assert light_model_class.number_light_planes == mass_model_class.number_mass_planes + 1
        self.MPMassModel = mass_model_class
        self.MPLightModel = light_model_class

        if source_grid_scale is None:
            self._source_grid_scale = [1.0] * self.MPLightModel.number_light_planes
        else:
            self._source_grid_scale = source_grid_scale

        if conjugate_points is None:
            self.conjugate_points = [None] * self.MPMassModel.number_mass_planes
        else:
            self.conjugate_points = conjugate_points

        for i, has_pixels in enumerate(self.MPMassModel.has_pixels):
            if has_pixels:
                pixel_grid = self.Grid.create_model_grid(
                    **self.MPMassModel.mass_models[i].pixel_grid_settings
                )
                self.MPMassModel.mass_models[i].set_pixel_grid(pixel_grid)

        for i, has_pixels in enumerate(self.MPLightModel.has_pixels):
            if has_pixels:
                pixel_grid = self.Grid.create_model_grid(
                    **self.MPLightModel.light_models[i].pixel_grid_settings
                )
                self.MPLightModel.light_models[i].set_pixel_grid(pixel_grid, self.Grid.pixel_area)
        if source_arc_masks is None:
            self.source_arc_masks = np.stack(
                [np.ones(self.Grid.num_pixel_axes)] * self.MPLightModel.number_light_planes
            )
        else:
            self.source_arc_masks = np.stack([
                np.ones(self.Grid.num_pixel_axes) if m is None else m for m in source_arc_masks
            ])

        self._src_adaptive_grid = self.MPLightModel.pixel_is_adaptive

        if kwargs_numerics is None:
            kwargs_numerics = {}
        self.ImageNumerics = Numerics(pixel_grid=self.Grid, psf=self.PSF, **kwargs_numerics)

        ssf = self.ImageNumerics.grid_supersampling_factor

        # get masks in super sampled space
        s_ones = np.ones([ssf, ssf])
        self.source_arc_masks_ss = np.stack([
            np.kron(m, s_ones) for m in self.source_arc_masks
        ])
        # flatten the super sampled masks
        self._source_arc_masks_flat = self.source_arc_masks_ss.reshape(
            self.MPLightModel.number_light_planes,
            -1
        )
        # plt.imshow(self._source_arc_masks_flat[0].reshape(self.Grid.num_pixel_axes), cmap='gray')
        # plt.show()
        # plt.imshow(self._source_arc_masks_flat[1].reshape(self.Grid.num_pixel_axes), cmap='gray')
        # plt.show()
        # get the (flattened) outline of the super sampled masks
        # these boundaries are used to define the extent of pixelated grids
        self._source_arc_masks_flat_bool = np.stack([
            (m - scipy.ndimage.binary_erosion(m))
            for m in self.source_arc_masks_ss
        ]).reshape(
            self.MPLightModel.number_light_planes,
            -1
        ).astype(bool)

    def k_extend(self, k, amount):
        if k is None:
            return jnp.arange(amount)
        elif isinstance(k, int):
            return jnp.array([k])
        else:
            return jnp.array(k)

    @partial(jax.jit, static_argnums=(0, 4, 5, 6, 7, 8, 9))
    def model(
        self,
        eta_flat=None,
        kwargs_mass=None,
        kwargs_light=None,
        supersampled=False,
        unconvolved=False,
        k_mass=None,
        k_light=None,
        k_planes=None,
        return_pixel_scale=False,
    ):
        '''Create the 2D model image from the parameter values.  Note: due to JIT compilation,
        the first call to this method will be slower.

        Parameters
        ----------
        eta_flat : jax.numpy array
            upper triangular elements of eta matrix, values defined as
            eta_ij = D_ij D_i+1 / D_j D_ii+1 where D_ij is the angular diameter
            distance between redshifts i and j. Only include values where
            j > i+1. This convention implies that all einstein radii are defined
            with respect to the **next** mass plane back (**not** the last plane in
            the stack).
            For more details, see comments in the Pull Request #37:
            https://github.com/Herculens/herculens/pull/37#issuecomment-2343179184
        kwargs_mass : list of list
            List of lists of parameter dictionaries of lens mass model parameters
            corresponding to each mass plane.
        kwargs_light : list of list
            List of lists of parameter dictionaries corresponding to each light plane.
        supersampled : bool, optional
            If True, returns the unconvolved model on the higher resolution grid, by default False
        unconvolved : bool, optional
            If True, does perform convolution with the PSF, by default False
        k_mass : list of list, optional
            Only evaluate the k-th mass model (list of list of index values) for each mass
            plane, by default None
        k_light : list of list, optional
            Only evaluate the k-th light model (list of list of index values) for each light
            plane, by default None
        k_planes : list, optional
            List of light plane index values to include in the output, by default None
        return_pixel_scale : bool, optional
            If True returns the pixel scale (arcsec/pixel) of each source plane, by default False.
            Note: requites and pixelated adaptive source grid to be used.

        Returns
        -------
        model : jax.numpy array
            The 2D model image for the lens system
        pixel_scale : list, optional
            The pixel scale (arcsec/pixel) of each source plane, by default False.
            Note: requires pixelated adaptive source grid to be used and have `return_pixel_scale`
            set to True.
        '''
        ra_grid_img, dec_grid_img = self.ImageNumerics.coordinates_evaluate

        # pixel grid positions on each mass plane (including the lens plane)
        ra_grid_planes, dec_grid_planes = self.MPMassModel.ray_shooting(
            ra_grid_img,
            dec_grid_img,
            eta_flat,
            kwargs_mass,
            k=k_mass
        )
        # (masked) light contribution from each plane
        pixels_x_coord, pixels_y_coord, _ = self.adapt_source_coordinates(
            ra_grid_planes,
            dec_grid_planes
        )
        light_planes = self.MPLightModel.surface_brightness(
            ra_grid_planes,
            dec_grid_planes,
            kwargs_light,
            pixels_x_coord,
            pixels_y_coord,
            k=k_light,
        ) * self._source_arc_masks_flat
        k_planes = self.k_extend(k_planes, len(light_planes))
        model = light_planes[k_planes].sum(axis=0)
        if not supersampled:
            model = self.ImageNumerics.re_size_convolve(model, unconvolved=unconvolved)
        if return_pixel_scale:
            pixel_scale = [x[1] - x[0] if x is not None else None for x in pixels_x_coord]
            return model, pixel_scale
        else:
            return model

    def simulation(
        self,
        add_poisson=True,
        add_gaussian=True,
        compute_true_noise_map=True,
        noise_seed=18,
        **model_kwargs
    ):
        """
        same as model() but with noise added

        :param compute_true_noise_map: if True (default), define the noise map (diagonal covariance matrix)
        to be the 'true' one, i.e. based on the noiseless model image.
        :param noise_seed: the seed that will be used by the PRNG from JAX to fix the noise realization.
        The default is the arbitrary value 18, so it is the user task to change it for different realizations.
        """
        if self.Noise is None:
            raise ValueError("Impossible to generate noise realization because no noise class has been set")
        model = self.model(**model_kwargs)
        noise = self.Noise.realisation(
            model,
            noise_seed,
            add_poisson=add_poisson,
            add_gaussian=add_gaussian
        )
        simu = model + noise
        self.Noise.set_data(simu)
        if compute_true_noise_map is True:
            self.Noise.compute_noise_map_from_model(model)
        return simu
    
    def C_D_model(self, model, **kwargs_noise):
        return self.Noise.C_D_model(model, **kwargs_noise)

    def normalized_residuals(self, data, model, kwargs_noise=None, mask=None):
        """
        compute the map of normalized residuals,
        given the data and the model image
        """
        if kwargs_noise is None:
            kwargs_noise = {}
        if mask is None:
            mask = np.ones(self.Grid.num_pixel_axes)
        noise_var = self.C_D_model(model, **kwargs_noise)
        noise = np.sqrt(noise_var)
        norm_res_model = (data - model) / noise * mask
        norm_res_tot = norm_res_model
        if mask is not None:
            # outside the mask just add pure data
            norm_res_tot += (data / noise) * (1. - mask)
        # make sure there is no NaN or infinite values
        norm_res_model = np.where(np.isfinite(norm_res_model), norm_res_model, 0.)
        norm_res_tot = np.where(np.isfinite(norm_res_tot), norm_res_tot, 0.)
        return norm_res_model, norm_res_tot

    def reduced_chi2(self, data, model, kwargs_noise=None, mask=None):
        """
        compute the reduced chi2 of the data given the model
        """
        if mask is None:
            mask = np.ones(self.Grid.num_pixel_axes)
        norm_res, _ = self.normalized_residuals(
            data, model, kwargs_noise=kwargs_noise, mask=mask
        )
        num_data_points = np.sum(mask)
        return np.sum(norm_res**2) / num_data_points

    @partial(jax.jit, static_argnums=(0, 3, 4))
    def trace_conjugate_points(self, eta_flat, kwargs_mass, N=1, k_mass=None):
        '''
        Helper function that can be used to ray-trace the list of conjugate points
        provided to the class on initialization to their corresponding source planes.
        '''
        i = N - 1
        if self.conjugate_points[i] is not None:
            x, y = self.conjugate_points[i].T
            conj_x, conj_y = self.MPMassModel.ray_shooting(
                x, y,
                eta_flat,
                kwargs_mass,
                k=k_mass,
                N=N,
            )
            return jnp.vstack([conj_x[-1], conj_y[-1]]).T
        else:
            return None

    def mask_extent(self, x_grid_src, y_grid_src, npix_src, source_grid_scale):
        '''Calculate the extent of an arc mask in it's source plane.

        Parameters
        ----------
        x_grid_src : jax.numpy array
            x positions of the arc mask in it's source plane
        y_grid_src : jax.numpy array
            y positions of the arc mask in it's source plane
        npix_src : int
            Number of pixels in the source plane
        source_grid_scale : float
            A float between 0 and 1 indicating a global scale factor for the grid extent
            (e.g. 0.5 will set the pixel grid to be 50% the extent defined by the arc mask
            traced back to the source plane)

        Returns
        -------
        x_adapt : jax.numpy array
            Positions of each x-pixel in the source plane corresponding to the smallest
            square grid that contains the arc mask with the result scaled by `source_grid_scale`.
        y_adapt : jax.numpy array
            Positions of each y-pixel in the source plane corresponding to the smallest
            square grid that contains the arc mask with the result scaled by `source_grid_scale`.
        extent : list
            The bounds of the adaptive grid
        '''
        # create grid encompassed by ray-traced coordinates
        x_left, x_right = x_grid_src.min(), x_grid_src.max()
        y_bottom, y_top = y_grid_src.min(), y_grid_src.max()
        # center of the region
        cx = 0.5 * (x_left + x_right)
        cy = 0.5 * (y_bottom + y_top)
        # get the width and height
        width = jnp.abs(x_left - x_right)
        height = jnp.abs(y_bottom - y_top)
        # choose the largest of the two to end up with a square region
        half_size = source_grid_scale * 0.5 * jnp.maximum(height, width)
        # recompute the new boundaries
        x_left = cx - half_size
        x_right = cx + half_size
        y_bottom = cy - half_size
        y_top = cy + half_size
        x_adapt = jnp.linspace(x_left, x_right, npix_src)
        y_adapt = jnp.linspace(y_bottom, y_top, npix_src)
        extent = [x_adapt[0], x_adapt[-1], y_adapt[0], y_adapt[-1]]
        return x_adapt, y_adapt, extent

    @partial(jax.jit, static_argnums=(0, 3, 4, 5))
    def adapt_source_coordinates(
        self,
        ra_grid_planes,
        dec_grid_planes,
        force=False,
        npix_src=100,
        source_grid_scale=1
    ):
        '''Calculate the extent of all arc mask in each of their source planes.

        Parameters
        ----------
        ra_grid_planes, dec_grid_planes : jax.numpy array
            All observed (supersampled) grid traced back to each plane (this will be masked down
            to each arc mask internally to avoid needing to re-calculate the ray shooting).
        force : bool, optional
            If True calculate the adaptive grid position even if no adaptive pixel grids
            are being used in the current model (useful for plotting), by default False.
        npix_src : int, optional
            The size of the adaptive grid to make, by default 100
        source_grid_scale : float, optional
            A float between 0 and 1 indicating a global scale factor for the grid extent
            (e.g. 0.5 will set the pixel grid to be 50% the extent defined by the arc mask
            traced back to the source plane), by default 1.

        Returns
        -------
        x_adapt : jax.numpy array
            Positions of each x-pixel in the source plane corresponding to the smallest
            square grid that contains the arc mask with the result scaled by `source_grid_scale`.
        y_adapt : jax.numpy array
            Positions of each y-pixel in the source plane corresponding to the smallest
            square grid that contains the arc mask with the result scaled by `source_grid_scale`.
        extent : list
            The bounds of the adaptive grid
        '''
        x_adapt = []
        y_adapt = []
        extent_adapt = []
        for i, adapt in enumerate(self._src_adaptive_grid):
            if adapt or force:
                if not force:
                    npix_src, _ = self.MPLightModel.light_models[i].pixel_grid.num_pixel_axes
                    grid_scale = self._source_grid_scale[i]
                else:
                    grid_scale = source_grid_scale
                x_adapt_i, y_adapt_i, extent_i = self.mask_extent(
                    ra_grid_planes[i][self._source_arc_masks_flat_bool[i]],
                    dec_grid_planes[i][self._source_arc_masks_flat_bool[i]],
                    npix_src,
                    grid_scale
                )
                x_adapt.append(x_adapt_i)
                y_adapt.append(y_adapt_i)
                extent_adapt.append(extent_i)
            else:
                x_adapt.append(None)
                y_adapt.append(None)
                extent_adapt.append(None)
        return x_adapt, y_adapt, extent_adapt

    def get_source_coordinates(
        self,
        eta_flat,
        kwargs_mass,
        force=False,
        npix_src=100,
        source_grid_scale=1.0
    ):
        '''Calculate the adaptive source coordinates give `eta_flat` and `kwargs_mass`.

        Parameters
        ----------
        eta_flat : jax.numpy array
            upper triangular elements of eta matrix, values defined as
            eta_ij = D_ij D_i+1 / D_j D_ii+1 where D_ij is the angular diameter
            distance between redshifts i and j. Only include values where
            j > i+1. This convention implies that all einstein radii are defined
            with respect to the **next** mass plane back (**not** the last plane in
            the stack).
        kwargs_mass : list of list
            List of lists of parameter dictionaries of lens mass model parameters
            corresponding to each mass plane.
        force : bool, optional
            If True calculate the adaptive grid position even if no adaptive pixel grids
            are being used in the current model (useful for plotting), by default False.
        npix_src : int, optional
            The size of the adaptive grid to make, by default 100
        source_grid_scale : float, optional
            A float between 0 and 1 indicating a global scale factor for the grid extent
            (e.g. 0.5 will set the pixel grid to be 50% the extent defined by the arc mask
            traced back to the source plane), by default 1.

        Returns
        -------
        x_adapt : jax.numpy array
            Positions of each x-pixel in the source plane corresponding to the smallest
            square grid that contains the arc mask with the result scaled by `source_grid_scale`.
        y_adapt : jax.numpy array
            Positions of each y-pixel in the source plane corresponding to the smallest
            square grid that contains the arc mask with the result scaled by `source_grid_scale`.
        extent : list
            The bounds of the adaptive grid
        '''
        ra_grid_img, dec_grid_img = self.ImageNumerics.coordinates_evaluate
        ra_grid_planes, dec_grid_planes = self.MPMassModel.ray_shooting(
            ra_grid_img,
            dec_grid_img,
            eta_flat,
            kwargs_mass
        )
        return self.adapt_source_coordinates(
            ra_grid_planes,
            dec_grid_planes,
            force=force,
            npix_src=npix_src,
            source_grid_scale=source_grid_scale
        )




# Defines a pixelated profile in the lens potential
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'austinpeel', 'aymgal'


import numpy as np
import jax
import jax.numpy as jnp
from jax import grad, jacfwd, jacrev, vmap
# from functools import partial

from utax.interpolation import BilinearInterpolator, BicubicInterpolator

from herculens.Util import util


__all__ = ['PixelatedPotential', 'PixelatedPotentialDirac', 'PixelatedFixed']


class PixelatedPotential(object):
    param_names = ['pixels']
    lower_limit_default = {'pixels': -1e10}
    upper_limit_default = {'pixels': 1e10}
    fixed_default = {key: False for key in param_names}

    _interp_types = ['fast_bilinear', 'bicubic']
    _deriv_types = ['interpol', 'autodiff']

    def __init__(self, interpolation_type='fast_bilinear', derivative_type='autodiff'):
        """Lensing potential on a fixed coordinate grid."""
        if interpolation_type not in self._interp_types:
            raise ValueError(f"Invalid interpolated mode ('{interpolation_type}'). Must be in {self._interp_types}.")
        if derivative_type not in self._deriv_types:
            raise ValueError(f"Unknown derivatives mode '{derivative_type}' "
                             f"(supported types are {self._deriv_types}).")
        self._deriv_type = derivative_type
        self._interp_class = None
        self._interp_type = interpolation_type
        if self._interp_type == 'fast_bilinear':
            try:
                from jaxinterp2d import CartesianGrid
            except ImportError as e:
                warnings.warn("jaxinterp2d is not installed for 'fast_bilinear' option; fall back on the standard 'bicubic' option.")
                self._interp_type = 'bicubic'
            else:
                self._interp_class = CartesianGrid
        if self._interp_class is None:
            if self._interp_type == 'bicubic':
                from utax.interpolation import BicubicInterpolator
                self._interp_class = BicubicInterpolator
        
        self._pixel_grid = None
        self._x_coords, self._y_coords = None, None

    @property
    def pixel_grid(self):
        return self._pixel_grid

    def function(self, x, y, pixels=None):
        if self._interp_type == 'fast_bilinear':
            f = self._function_fast(x, y, pixels)
        elif self._interp_type == 'bicubic':
            f = self._function_std(x, y, pixels)
        return f

    def _function_fast(self, x, y, pixels):
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x.flatten(), y.flatten())
        x_, y_ = x_.reshape(*x.shape), y_.reshape(*y.shape)
        interp = self._interp_class(self._limits, pixels, cval=0.)
        return interp(y_, x_)

    def _function_std(self, x, y, pixels):
        """Interpolated evaluation of a pixelated light profile.

        Parameters
        ----------
        x, y : array-like
            Coordinates at which to evaluate the light profile.
        x_coords : 1D array
            Rectangular x-coordinate grid values.
        y_coords : 1D array
            Rectangular y-coordinate grid values.
        pixels : 2D array
            Surface brightness at fixed coordinate grid positions.
        method : str
            Interpolation method, either 'bilinear' or 'bicubic'.

        """
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x.flatten(), y.flatten())
        x_, y_ = x_.reshape(*x.shape), y_.reshape(*y.shape)
        # setup interpolation, assuming cartesian grid
        # Due to matching scipy's interpolation, we need to switch x and y
        # coordinates as well as transpose
        interp = self._interp_class(self._y_coords, self._x_coords, pixels)
        # evaluate the interpolator
        return interp(y_, x_)

    def derivatives(self, x, y, pixels):
        """Spatial first derivatives of the lensing potential.

        Parameters
        ----------
        x, y : array-like
            Coordinates at which to evaluate the lensing potential derivatives.
        pixels : 2D array
            Values of the lensing potential at fixed coordinate grid positions.

        """
        if self._deriv_type == 'interpol':
            return self._derivatives_interpol(x, y, pixels)
        elif self._deriv_type == 'autodiff':
            return self._derivatives_autodiff(x, y, pixels)

    def _derivatives_interpol(self, x, y, pixels):
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x.flatten(), y.flatten())
        x_, y_ = x_.reshape(*x.shape), y_.reshape(*y.shape)
        if 'bilinear' in self._interp_type:
            interp = BilinearInterpolator(self._y_coords, self._x_coords, pixels)
        else:
            interp = BicubicInterpolator(self._y_coords, self._x_coords, pixels)
        f_x = interp(y_, x_, dy=1) / self._delta_x
        f_y = interp(y_, x_, dx=1) / self._delta_y
        return f_x, f_y

    def _derivatives_autodiff(self, x, y, pixels):
        def function(params):
            res = self.function(params[0], params[1], pixels)
            if self._interp_type != 'fast_bilinear':
                res = res[0]
            return res
        grad_func = grad(function)
        param_array = jnp.array([x.flatten(), y.flatten()]).T
        res = vmap(grad_func)(param_array)
        f_x = res[:, 0].reshape(*x.shape)
        f_y = res[:, 1].reshape(*x.shape)
        return f_x, f_y

    def hessian(self, x, y, pixels):
        """Spatial second derivatives of the lensing potential.

        Parameters
        ----------
        x, y : array-like
            Coordinates at which to evaluate the lensing potential derivatives.
        pixels : 2D array
            Values of the lensing potential at fixed coordinate grid positions.

        """
        if self._deriv_type == 'interpol':
            return self._hessian_interpol(x, y, pixels)
        elif self._deriv_type == 'autodiff':
            return self._hessian_autodiff(x, y, pixels)

    def _hessian_interpol(self, x, y, pixels):
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x.flatten(), y.flatten())
        x_, y_ = x_.reshape(*x.shape), y_.reshape(*y.shape)
        # second-oder derivatives can only be obtained with bicubic interpolation (or higher orders)
        interp = BicubicInterpolator(self._y_coords, self._x_coords, pixels)
        f_xx = interp(y_, x_, dy=2) / self._delta_x**2
        f_yy = interp(y_, x_, dx=2) / self._delta_y**2
        f_xy = interp(y_, x_, dx=1, dy=1) /self._delta_x*self._delta_y
        return f_xx, f_yy, f_xy

    def _hessian_autodiff(self, x, y, pixels):
        def function(params):
            res = self.function(params[0], params[1], pixels)
            if self._interp_type != 'fast_bilinear':
                res = res[0]
            return res
        hessian_func = jacfwd(jacrev(function))
        param_array = jnp.array([x.flatten(), y.flatten()]).T
        res = vmap(hessian_func)(param_array)
        f_xx = res[:, 0, 0].reshape(*x.shape)
        f_xy = res[:, 0, 1].reshape(*x.shape)
        # f_yx = res[:, 1, 0].reshape(*x.shape)
        f_yy = res[:, 1, 1].reshape(*x.shape)
        return f_xx, f_yy, f_xy

    def set_pixel_grid(self, pixel_grid):
        self._pixel_grid = pixel_grid
        # compute the original pixel size (used for proper scaling of derivatives)
        x_grid, y_grid = self.pixel_grid.pixel_coordinates
        self._delta_x = abs(x_grid[0, 1] - x_grid[0, 0])
        self._delta_y = abs(y_grid[1, 0] - y_grid[0, 0])
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_grid, y_grid = self.pixel_grid.map_coord2pix(util.image2array(x_grid), util.image2array(y_grid))
        x_grid, y_grid = util.array2image(x_grid), util.array2image(y_grid)
        self._x_coords = x_grid[0, :]
        self._y_coords = y_grid[:, 0]
        self._limits = [
            ( self._y_coords.min(), self._y_coords.max() ), 
            ( self._x_coords.min(), self._x_coords.max() )
        ]



class PixelatedPotentialDirac(object):
    param_names = ['psi', 'center_x', 'center_y']
    lower_limit_default = {'psi': -1e10, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'psi': 1e10, 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}

    def __init__(self):
        """Dirac impulse in potential on a fixed coordinate grid."""
        super(PixelatedPotentialDirac, self).__init__()
        self.pp = PixelatedPotential()

    @property
    def pixel_grid(self):
        return self.pp.pixel_grid

    def function(self, x, y, psi, center_x, center_y):
        """Interpolated evaluation of the lensing potential.

        Parameters
        ----------
        x, y : array-like
            Coordinates at which to evaluate the lensing potential.
        psi : float
            Value of the pixelated lensing potential at x, y
        center_x : float
            center in x-coordinate
        center_y : float
            center in y-coordinate

        """
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x, y)
        return jnp.where((x_ >= center_x - self.hss_x) & 
                         (x_ <= center_x + self.hss_x) & 
                         (y_ >= center_y - self.hss_y) & 
                         (y_ <= center_y + self.hss_y), 
                         psi, 0.)

    def derivatives(self, x, y, psi, center_x, center_y):
        """Spatial first derivatives of the lensing potential.

        Parameters
        ----------
        x, y : array-like
            Coordinates at which to evaluate the lensing potential derivatives.
        psi : 2D array
            Values of the lensing potential at fixed coordinate grid positions.
        center_x : float
            center in x-coordinate
        center_y : float
            center in y-coordinate

        """
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x, y)
        # the following array is at the input (x, y) resolution
        rect_shape = int(np.sqrt(x_.size)), int(np.sqrt(y_.size))
        pixels = self.function(x_, y_, psi, center_x, center_y).reshape(rect_shape)
        # we the want to interpolate it to the resolution of the underlying pixelated grid
        # first, get the axes of from input coordinates
        x_coords, y_coords = np.reshape(x_, rect_shape)[0, :], np.reshape(y_, rect_shape)[:, 0]
        # create the interpolator
        interp = BicubicInterpolator(y_coords, x_coords, pixels)
        # define the 2D coordinate underlying grid
        x_grid, y_grid = np.meshgrid(self.pp.x_coords, self.pp.y_coords)
        # the result is the pixels interpolated on the same grid as the underlying pixelated profile
        pixels_grid = interp(y_grid, x_grid)
        # call the derivatives on the underlying pixelated profile
        return self.pp.derivatives(x_, y_, pixels_grid)

    def hessian(self, x, y, psi, center_x, center_y):
        """Spatial second derivatives of the lensing potential.

        Parameters
        ----------
        x, y : array-like
            Coordinates at which to evaluate the lensing potential derivatives.
        psi : 2D array
            Values of the lensing potential at fixed coordinate grid positions.
        center_x : float
            center in x-coordinate
        center_y : float
            center in y-coordinate

        """
        raise NotImplementedError("Computation of Hessian terms for PixelatedPotentialDirac is not implemented.")

    def set_pixel_grid(self, pixel_grid):
        self.pp.set_pixel_grid(pixel_grid)
        self.hss_x = np.abs(self.pp._x_coords[0] - self.pp._x_coords[1]) / 2.
        self.hss_y = np.abs(self.pp._y_coords[0] - self.pp._y_coords[1]) / 2.



class PixelatedFixed(object):
    # param_names = ['pot_pixels',
    #                'deriv_pixels',
    #                'hess_pixels']
    # lower_limit_default = {'pixels': -1e10}
    # upper_limit_default = {'pixels': 1e10}
    # fixed_default = {key: False for key in param_names}
    def __init__(self, func_pixel_grid=None, deriv_pixel_grid=None, hess_pixel_grid=None):
        """
        Potential and its spatial derivatives defined on a pixelated grid (can be different for each of these).
        IMPORTANT: This profile is NOT designed to be used during optimization!!
        """
        self._func_pixel_grid  = func_pixel_grid
        self._deriv_pixel_grid = deriv_pixel_grid
        self._hess_pixel_grid  = hess_pixel_grid
        try:
            from jaxinterp2d import CartesianGrid
        except ImportError as e:
            raise NotImplementedError("`jaxinterp2d` is not installed and "
                                      "the use of the bilinear interpolator from `utax` is not implemented.")
            # self._interp_type = 'bilinear'
            # self._interp_class = BilinearInterpolator
        else:
            self._interp_type = 'fast_bilinear'
            self._interp_class = CartesianGrid

    def get_pixel_grid(self, component='func_pixels'):
        if component == 'func_pixels':
            return self._func_pixel_grid
        elif component == 'deriv_pixels':
            return self._deriv_pixel_grid
        elif component == 'hess_pixels':
            return self._hess_pixel_grid
        else:
            raise ValueError(f"Invalid component '{component}' for pixelated mass profile.")

    def function(self, x, y, func_pixels=None, deriv_pixels=None, hess_pixels=None):
        return self._interpol(x, y, func_pixels, self._func_pixel_grid)
    
    def derivatives(self, x, y, func_pixels=None, deriv_pixels=None, hess_pixels=None):
        deriv_x, deriv_y = deriv_pixels
        return (self._interpol(x, y, deriv_x, self._deriv_pixel_grid), 
                self._interpol(x, y, deriv_y, self._deriv_pixel_grid))

    def hessian(self, x, y, func_pixels=None, deriv_pixels=None, hess_pixels=None):
        hess_xx, hess_yy, hess_xy = hess_pixels
        return (self._interpol(x, y, hess_xx, self._hess_pixel_grid), 
                self._interpol(x, y, hess_yy, self._hess_pixel_grid),
                self._interpol(x, y, hess_xy, self._hess_pixel_grid))

    def _interpol(self, x, y, pixels, pixel_grid):
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = pixel_grid.map_coord2pix(x.flatten(), y.flatten())
        x_, y_ = x_.reshape(*x.shape), y_.reshape(*y.shape)
        x_coord, y_coord = pixel_grid.pixel_axes
        x_coord_pix, y_coord_pix = pixel_grid.map_coord2pix(x_coord, y_coord)
        limits = [
            ( y_coord_pix.min(), y_coord_pix.max() ), 
            ( x_coord_pix.min(), x_coord_pix.max() )
        ]
        interp = self._interp_class(limits, pixels, cval=0.)
        return interp(y_, x_)



# Defines a pixelated profile
# 
# Copyright (c) 2023, herculens developers and contributors

__author__ = 'austinpeel', 'aymgal'


import warnings
import numpy as np
import jax.numpy as jnp
from jax import grad, jacfwd, jacrev, vmap

from utax.interpolation import BilinearInterpolator, BicubicInterpolator

from herculens.Util import util


__all__= ['Pixelated']


class Pixelated(object):

    """
    Surface brightness defined on a fixed coordinate grid.
    Note: 'fast_bilinear' is only marginally faster than 'bilinear'.
    """

    param_names = ['pixels']
    lower_limit_default = {'pixels': -1e10}
    upper_limit_default = {'pixels': 1e10}
    fixed_default = {key: False for key in param_names}
    
    _interp_types = ['fast_bilinear', 'bilinear', 'bicubic']
    _deriv_types = ['interpol', 'autodiff']

    def __init__(self, interpolation_type='fast_bilinear', allow_extrapolation=True, 
                 derivative_type='interpol', adaptive_grid=False):
        if interpolation_type not in self._interp_types:
            raise ValueError(f"Invalid method ('{interpolation_type}'). Must be in {self._interp_types}.")
        if derivative_type not in self._deriv_types:
            raise ValueError(f"Unknown derivatives mode '{derivative_type}' "
                             f"(supported types are {self._deriv_types}).")
        self._deriv_type = derivative_type
        self._interp_type = interpolation_type
        self._interp_class = None
        if self._interp_type == 'fast_bilinear':
            try:
                from jaxinterp2d import CartesianGrid
            except ImportError as e:
                warnings.warn("jaxinterp2d is not installed for 'fast_bilinear' option; fall back on the standard 'bilinear' option.")
                self._interp_type = 'bilinear'
            else:
                self._interp_class = CartesianGrid
        if self._interp_class is None:
            if self._interp_type == 'bilinear':
                self._interp_class = BilinearInterpolator
            elif self._interp_type == 'bicubic':
                self._interp_class = BicubicInterpolator
        self._extrapol_bool = allow_extrapolation
        self._adaptive_grid = adaptive_grid

        self._data_pixel_area = None
        self._pixel_grid = None
        self._x_coords, self._y_coords = None, None

    @property
    def num_amplitudes(self):
        if self._x_coords is None:
            raise ValueError("No coordinates axes have been set for pixelated profile.")
        return self._x_coords.size

    @property
    def pixel_grid(self):
        return self._pixel_grid
    
    @property
    def is_adaptive(self):
        return self._adaptive_grid

    def function(self, x, y, pixels_x_coord=None, pixels_y_coord=None, pixels=None):
        if self._interp_type == 'fast_bilinear':
            f = self._function_fast(x, y, pixels_x_coord, pixels_y_coord, pixels)
        elif self._interp_type in ['bilinear', 'bicubic']:
            f = self._function_std(x, y, pixels_x_coord, pixels_y_coord, pixels)
        # normalize for correct units when evaluated by LensImage methods
        return f / self._data_pixel_area

    def _function_fast(self, x, y, pixels_x_coord, pixels_y_coord, pixels):
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x.flatten(), y.flatten())
        x_, y_ = x_.reshape(*x.shape), y_.reshape(*y.shape)
        if self._adaptive_grid:
            x_coord_pix, y_coord_pix = self.pixel_grid.map_coord2pix(pixels_x_coord, pixels_y_coord)
            limits = [
                ( y_coord_pix.min(), y_coord_pix.max() ), 
                ( x_coord_pix.min(), x_coord_pix.max() )
            ]
        else:
            limits = self._limits
        interp = self._interp_class(limits, pixels, cval=0.)
        f = interp(y_, x_)
        return f

    def _function_std(self, x, y, pixels_x_coord, pixels_y_coord, pixels):
        """Interpolated evaluation of a pixelated light profile.

        Parameters
        ----------
        x, y : array-like
            Coordinates at which to evaluate the light profile.
        x_coords : 1D array
            Rectangular x-coordinate grid values.
        y_coords : 1D array
            Rectangular y-coordinate grid values.
        pixels : 2D array
            Surface brightness at fixed coordinate grid positions.
        method : str
            Interpolation method, either 'bilinear' or 'bicubic'.

        """
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x.flatten(), y.flatten())
        x_, y_ = x_.reshape(*x.shape), y_.reshape(*y.shape)
        # setup interpolation, assuming cartesian grid
        if not self._adaptive_grid:  # in this case pixels_x_coord and pixels_y_coord should be None
            pixels_x_coord, pixels_y_coord = self._x_coords, self._y_coords
        interp = self._interp_class(pixels_y_coord, pixels_x_coord, pixels,
                                    allow_extrapolation=self._extrapol_bool)
        # evaluate the interpolator
        f = interp(y_, x_)
        return f

    def derivatives(self, x, y, pixels_x_coord=None, pixels_y_coord=None, pixels=None):
        """Spatial first derivatives of the pixelated light profile.

        Parameters
        ----------
        x, y : array-like
            Coordinates at which to evaluate the surface brightness derivatives.
        pixels : 2D array
            Values of the surface brightness at fixed coordinate grid positions (surf. bright. units / arcsec)

        """
        if self._deriv_type == 'interpol':
            f_x, f_y = self._derivatives_interpol(x, y, pixels_x_coord, pixels_y_coord, pixels)
        elif self._deriv_type == 'autodiff':
            f_x, f_y = self._derivatives_autodiff(x, y, pixels_x_coord, pixels_y_coord, pixels)
        # normalize for correct units when evaluated by LensImage methods
        return f_x / self._data_pixel_area, f_y / self._data_pixel_area

    def _derivatives_interpol(self, x, y, pixels_x_coord, pixels_y_coord, pixels):
        if self._interp_type not in ['bilinear', 'bicubic']:
            raise ValueError(f"Invalid interpolation type '{self._interp_type}' for "
                             f"'interpolated' derivatives computation. "
                             f"Either use 'autodiff' derivatives computation, "
                             f"or choose interpolation type 'bilinear' or 'bicubic'.")
        elif self._interp_type == 'bilinear':
            interp_class = BilinearInterpolator
        else:
            interp_class = BicubicInterpolator
        # ensure the coordinates are cartesian by converting angular to pixel units
        x_, y_ = self.pixel_grid.map_coord2pix(x.flatten(), y.flatten())
        x_, y_ = x_.reshape(*x.shape), y_.reshape(*y.shape)
        # setup interpolation, assuming cartesian grid
        if not self._adaptive_grid:  # in this case pixels_x_coord and pixels_y_coord should be None
            pixels_x_coord, pixels_y_coord = self._x_coords, self._y_coords
        interp = interp_class(pixels_y_coord, pixels_x_coord, pixels,
                              allow_extrapolation=self._extrapol_bool)
        # evaluate the interpolator
        f_x = interp(y_, x_, dy=1) / self._delta_x
        f_y = interp(y_, x_, dx=1) / self._delta_y
        return f_x, f_y

    def _derivatives_autodiff(self, x, y, pixels_x_coord, pixels_y_coord, pixels):
        def function(params):
            res = self.function(
                params[0], params[1], 
                pixels_x_coord=pixels_x_coord, pixels_y_coord=pixels_y_coord,
                pixels=pixels)
            if self._interp_type != 'fast_bilinear':
                res = res[0]
            return res
        grad_func = grad(function)
        param_array = jnp.array([x.flatten(), y.flatten()]).T
        res = vmap(grad_func)(param_array)
        f_x = res[:, 0].reshape(*x.shape)
        f_y = res[:, 1].reshape(*x.shape)
        return f_x, f_y
    
    def set_pixel_grid(self, pixel_grid, data_pixel_area):
        self._data_pixel_area = data_pixel_area
        self._pixel_grid = pixel_grid
        # compute the original pixel size (used for proper scaling of derivatives)
        x_grid, y_grid = self.pixel_grid.pixel_coordinates
        self._delta_x = abs(x_grid[0, 1] - x_grid[0, 0])
        self._delta_y = abs(y_grid[1, 0] - y_grid[0, 0])
        # ensure the coordinates are cartesian by converting angular to pixel units
        nx, ny = x_grid.shape
        x_grid, y_grid = self.pixel_grid.map_coord2pix(util.image2array(x_grid), 
                                                       util.image2array(y_grid))
        self._x_coords = util.array2image(x_grid, nx=nx, ny=ny)[0, :]
        self._y_coords = util.array2image(y_grid, nx=nx, ny=ny)[:, 0]
        self._limits = [
            ( self._y_coords.min(), self._y_coords.max() ), 
            ( self._x_coords.min(), self._x_coords.max() )
        ]
        



# Utility functions to generate gaussian random fields (GRF)
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'gvernard', 'aymgal'


import math
import numpy as np
from numpy.fft import fft, fft2, fftn, fftfreq, ifft, fftshift

#################################################################################################
# Power spectrum
#################################################################################################

def grf_normalization(nx, ny, Lx, Ly, power):
    Psum = Psum_calculator(nx, ny, Lx, Ly, power)
    Npix = nx*ny
    return (Npix**2.)/(2.*Psum)

def powspec(L, variance, Npix, Psum, power):
    """Parametrization of the power spectrum"""
    if L <= 0:
        P = 0.0 # variance*(Npix**2.)/(2.*Psum)*L**(0.0)
    else:
        A = variance*(Npix**2.)/(2.*Psum)
        P = A*L**(power)
    return P


def Psum_calculator(nx, ny, Lx, Ly, power):
    """Sum over the power spectrum"""
    #lxaxis = np.append (np.arange (0.,(nx/2.)/Lx, 1./Lx), np.arange ((-nx/2.)/Lx, 0.,1./Lx))
    #lyaxis = np.append (np.arange (0.,(ny/2.)/Ly, 1./Ly), np.arange ((-ny/2.)/Ly, 0.,1./Ly))
    resolution = float(Lx) / float(nx)
    lxaxis = np.fft.fftfreq (nx, resolution)
    lyaxis = np.fft.fftfreq (ny, resolution)

    lx = list (np.zeros([nx, 1]))
    ly = list (np.zeros([ny, 1]))

    for x in range (len(lx)):
        lx [x] = lxaxis
    
    for y in range (len(ly)):
        ly [y] = lyaxis
    
    lx = np.array (lx)
    ly = np.transpose(np.array (ly))
    l  = np.sqrt (lx**2. + ly**2.)
    
    summ = 0.
    for y in range(np.shape(l)[0]):
        for x in range(np.shape(l)[1]):
            if l[y][x]==0.:
                summ += 0.
            else:
                summ += l[y][x]**(power)
    return summ
    

def gauss_rand_2d(nx, ny, Lx, Ly, var, power, seed=None):
    """
    Generate a realization of a Gaussian Random Field (GRF) with variance `var` and slope `power`.
    
    nx: number of pixels along x
    ny: number of pixels along y
    Lx: physical length of the field along x (typically in arcsec)
    Ly: physical length of the field along y (typically in arcsec)
    var: approximated variance of the GRF over the field-of-view (the larger the field-of-view, the more precise it is)
    beta: slope of the power spectrum
    seed: number used as a seed for the PRNG (numpy.random.seed)
    """

    if seed is not None:
        np.random.seed(seed)

    j= 0 + 1j # Defining the complex number
    plane = np.zeros ([nx, ny], dtype='cfloat') # Empty matrix to be filled in for the Fourier plane

    #lxaxis = np.append (np.arange (0.,(nx/2.)/Lx, 1./Lx), np.arange ((-nx/2.)/Lx, 0.,1./Lx))
    #lyaxis = np.append (np.arange (0.,(ny/2.)/Ly, 1./Ly), np.arange ((-ny/2.)/Ly, 0.,1./Ly))

    resolution = float(Lx) / float(nx)
    lxaxis = np.fft.fftfreq (nx, resolution)
    lyaxis = np.fft.fftfreq (ny, resolution)

    Psum = Psum_calculator(nx, ny, Lx, Ly, power)

    for y in range(np.shape(plane)[0]):
        for x in range(np.shape(plane)[1]):
            # The coordinates centered at x = n/2, y = n/2
            i1 = x - nx/2
            j1 = y - ny/2 

            #The coordinates in the Fourier plane
            lx = lxaxis [x]
            ly = lyaxis [y]
            # The magnitude of the l-vector
            l = np.sqrt (lx**2. + ly**2.) 

            # Polar Box-Muller transform
            sigma = math.sqrt(powspec(l, var, nx*ny, Psum, power))
            s = 1.1
            while s > 1. :
                u = np.random.uniform (-1.,1.) 
                v = np.random.uniform (-1.,1.)
                s = u**2. + v**2.
            fac = math.sqrt(-2.*math.log(s)/s)
            z1 = u*fac*sigma
            z2 = v*fac*sigma

            # Filling in the grid
            if x==0 and y==0: # average of the field
                plane[y][x] = 0.0

            # three points that need to be real valued to get a real image after FFT:
            elif x== 0 and y==ny/2:
                plane[y][x] = z1
            elif x==nx/2 and y==0:
                plane[y][x] = z1
            elif x==nx/2 and y==ny/2:
                plane[y][x] = z1
            else :
                plane[y][x] = z1+j*z2

            # Creating symmetry f(k) = f*(-k)
            y2 = int( -(j1 + ny/2) )
            x2 = int( -(i1 + nx/2) )
        
            plane[y2][x2] = plane[y][x].conjugate()

        if y>np.shape(plane)[0]/2.:
            break
    return np.fft.ifftshift(np.fft.ifft2(plane)).real




from .info import version_info, __version__, __author__, __license__

from .Coordinates.pixel_grid import PixelGrid
from .Instrument.psf import PSF
from .Instrument.noise import Noise

# NOTE: non-elliptical profiles will be suppressed in the future
from .LightModel.light_model import LightModel
from .LightModel.light_model_multiplane import MPLightModel
from .LightModel.Profiles.gaussian import Gaussian as GaussianLight, GaussianEllipse as GaussianEllipseLight
from .LightModel.Profiles.multipole import Multipole
from .LightModel.Profiles.sersic import Sersic, SersicElliptic
from .LightModel.Profiles.shapelets import Shapelets
from .LightModel.Profiles.uniform import Uniform
from .LightModel.Profiles.pixelated import Pixelated as PixelatedLight

from .MassModel.mass_model import MassModel
from .MassModel.mass_model_multiplane import MPMassModel
from .MassModel.Profiles.sis import SIS  # NOTE: this will be suppressed in the future
from .MassModel.Profiles.sie import SIE
from .MassModel.Profiles.nie import NIE
from .MassModel.Profiles.epl import EPL
from .MassModel.Profiles.shear import Shear, ShearGammaPsi
from .MassModel.Profiles.gaussian_potential import Gaussian as GaussianPotential
from .MassModel.Profiles.point_mass import PointMass
from .MassModel.Profiles.multipole import Multipole
from .MassModel.Profiles.pixelated import (
    PixelatedPotential, PixelatedPotentialDirac, PixelatedFixed
)
from .MassModel.Profiles.piemd import PIEMD
from .MassModel.Profiles.dpie import DPIE

from .PointSourceModel.point_source_model import PointSourceModel

try:
    from .GenericModel.correlated_field import CorrelatedField
except ImportError:
    # an error will be raised via the from herculens.Util.jifty_util submodule, 
    # typically when instantiating the CorrelatedField class
    pass

from .LensImage.lens_image import LensImage, LensImage3D
from .LensImage.lens_image_multiplane import MPLensImage
from .Inference.loss import Loss
from .Inference.ProbModel.numpyro import NumpyroModel
from .Inference.Optimization.jaxopt import JaxoptOptimizer
from .Inference.Optimization.optax import OptaxOptimizer
from .Analysis.plot import Plotter

from .Util import param_util as prmu
from .Util import plot_util as pltu



# Defines en dual pseudo isothermal elliptical (dPIE) mass profile
#
# This is based on the GLEE definition of the profile (Suyu & Halkola 2010, Suyu et al. 2012).
# 
# Copyright (c) 2025: herculens developers and contributors


__author__ = 'aymgal'


import numpy as np
import jax.numpy as jnp
from functools import partial
from jax import jit
from jax.tree_util import register_pytree_node

from herculens.MassModel.Profiles.piemd import PIEMD


__all__ = [
    'DPIE', 
]


class DPIE(object):
    """
    Dual pseudo isothermal elliptical (dPIE) mass profile, based on the
    different of two PIEMD profiles as implemented in GLEE.

    The convergence is
    kappa(x,y) = (Elimit / 2) * (s^2/(s^2-w^2)) * (1/sqrt(w^2 + rem^2) - 1/sqrt(s^2 + rem^2)
    
    with parameters
    Elimit = strength, (= Einstein radius of SIS in limiting case of s->inf, w->0)
    w = core radius
    s = truncation/scale radius (that must be >w)
    rem = elliptical mass radius = sqrt(x^2/(1+e)^2 + y^2/(1-e)^2)
    e = ellipticity = (1-q)/(1+q)
    q = axis ratio (that is <=1)

    Parameters
    ----------
    r_soft : _type_, optional
        Softening radius, by default 1e-8
    scale_flag : bool, optional
        'scale' flag usually given in GLEE config files, by default True, 
        which leads to theta_E being simply the lensing strength.
    """
    param_names = ['theta_E', 'r_core', 'r_trunc', 'q', 'phi', 'center_x', 'center_y']
    lower_limit_default = {'theta_E': 0, 'r_core': 0, 'r_trunc': 0, 'q': 0.2, 'phi': -np.pi/2., 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'theta_E': 10, 'r_core': 1e10, 'r_trunc': 1e10, 'q': 1., 'phi': +np.pi/2., 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}
    
    def __init__(self, r_soft=1e-8, scale_flag=True):
        self._r_soft = r_soft
        self._scale_flag = scale_flag # if True, theta_E corresponds to the Einstein radius of the profile
        self._piemd = PIEMD(r_soft=self._r_soft, scale_flag=False)  # the PIEMD scale flag must be False here.
    
    @partial(jit, static_argnums=(0,))
    def function(self, x, y, theta_E, r_core, r_trunc, q, phi, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param r_core: core radius
        :param r_trunc: truncation radius
        :param q: axis ratio
        :param phi: position angle
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        theta_E_piemd, w, s = self._param_conv(theta_E, r_core, r_trunc, self._scale_flag)
        f_w = self._piemd.function(x, y, theta_E_piemd, w, q, phi, center_x=center_x, center_y=center_y)
        f_s = self._piemd.function(x, y, theta_E_piemd, s, q, phi, center_x=center_x, center_y=center_y)
        f = f_w - f_s
        return f

    @partial(jit, static_argnums=(0,))
    def derivatives(self, x, y, theta_E, r_core, r_trunc, q, phi, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param r_core: core radius
        :param r_trunc: truncation radius
        :param q: axis ratio
        :param phi: position angle
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        theta_E_piemd, w, s = self._param_conv(theta_E, r_core, r_trunc, self._scale_flag)
        f_x_w, f_y_w = self._piemd.derivatives(x, y, theta_E_piemd, w, q, phi, center_x=center_x, center_y=center_y)
        f_x_s, f_y_s = self._piemd.derivatives(x, y, theta_E_piemd, s, q, phi, center_x=center_x, center_y=center_y)
        f_x = f_x_w - f_x_s
        f_y = f_y_w - f_y_s
        return f_x, f_y
    
    @partial(jit, static_argnums=(0,))
    def hessian(self, x, y, theta_E, r_core, r_trunc, q, phi, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param r_core: core radius
        :param r_trunc: truncation radius
        :param q: axis ratio
        :param phi: position angle
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        theta_E_piemd, w, s = self._param_conv(theta_E, r_core, r_trunc, self._scale_flag)
        f_xx_w, f_yy_w, f_xy_w = self._piemd.hessian(x, y, theta_E_piemd, w, q, phi, center_x=center_x, center_y=center_y)
        f_xx_s, f_yy_s, f_xy_s = self._piemd.hessian(x, y, theta_E_piemd, s, q, phi, center_x=center_x, center_y=center_y)
        f_xx = f_xx_w - f_xx_s
        f_yy = f_yy_w - f_yy_s
        f_xy = f_xy_w - f_xy_s
        return f_xx, f_yy, f_xy
    
    def _param_conv(self, theta_E, r_core, r_trunc, scale_flag):
        """This functions follows the GLEE documentation"""
        w, s = r_core, r_trunc
        w, s = self._check_radii(w, s)
        if scale_flag is True:
            E0 = theta_E
            prefac = E0**2 / ( (jnp.sqrt(w**2 + E0**2) - w) - (jnp.sqrt(s**2 + E0**2) - s) )
        else:
            Elim = theta_E
            prefac = Elim * s**2 / (s**2 - w**2)
        return prefac, w, s
    
    def _check_radii(self, w, s):
        # make sure the core radius parameters do not go below some small value for numerical stability
        w = jnp.where(w < self._r_soft, self._r_soft, w)
        # NOTE: the following swap of values *may* cause issues with JAX autodiff
        w_ = jnp.where(s < w, s, w)
        s_ = jnp.where(s < w, w, s)
        return w_, s_

def flatten_func(self):
    children = (self._piemd,)
    aux_data = {
        'r_soft': self._r_soft,
        'scale_flag': self._scale_flag,
    }
    return (children, aux_data)

def unflatten_func(aux_data, children):
    # Here we avoid `__init__` because it has extra logic we don't require:
    obj = object.__new__(DPIE)
    obj._piemd = children[0]
    obj._r_soft = aux_data['r_soft']
    obj._flag_dpie = aux_data['scale_flag']
    return obj

register_pytree_node(DPIE, flatten_func, unflatten_func)



# Copyright (c) 2023, herculens developers and contributors

__author__ = 'austinpeel'

from herculens.PointSourceModel.point_source import PointSource

__all__ = ['PointSourceModel']

SUPPORTED_TYPES = ['IMAGE_POSITIONS', 'SOURCE_POSITION']


class PointSourceModel(object):
    """Collection of point sources defined in the source or image plane."""

    param_names = ['ra', 'dec', 'amp']

    def __init__(self, point_source_type_list, mass_model=None, image_plane=None):
        """Instantiate a point source model.

        Parameters
        ----------
        point_source_type_list : list of str
            List of point source types to model.
        mass_model : instance of `herculens.MassModel.mass_model.MassModel`
            Model of the lensing mass used to map positions between the source
            and image planes. Default is None.
        image_plane : instance of `herculens.Coordinates.pixel_grid.PixelGrid`
            Pixel grid used for triangulation in solving the lens equation.

        """
        self.point_sources = []

        # Validate inputs
        if type(point_source_type_list) is not list:
            raise ValueError("point_source_type_list must be a list")

        # Populate point source list
        for ps_type in point_source_type_list:
            if ps_type in SUPPORTED_TYPES:
                ps = PointSource(ps_type, mass_model, image_plane)
                self.point_sources.append(ps)
            else:
                err_msg = (f"{ps_type} is not a valid point source type. " +
                           f"Supported types include {SUPPORTED_TYPES}")
                raise ValueError(err_msg)

        self.type_list = point_source_type_list

    def get_multiple_images(self, kwargs_point_source, kwargs_lens=None,
                            kwargs_solver=None, k=None, with_amplitude=True,
                            zero_amp_duplicates=True, re_compute=False):
        """Compute point source positions and amplitudes in the image plane.

        For point sources defined in the source plane, solving the lens
        equation is necessary in order to compute the corresponding (multiple)
        image plane positions.

        Parameters
        ----------
        kwargs_point_source : list of dict
            Keyword arguments corresponding to the point source instances.
        kwargs_lens : list of dict, optional
            Keyword arguments for the lensing mass model. Default is None.
        kwargs_solver : dict, optional
            Keyword arguments for the lens equation solver. Default is None.
        k : int, optional
            Index of the single point source for which to compute positions.
            If None, compute positions for all point sources.
        with_amplitude : bool, optional
            Whether to return the (magnified) amplitude of each point source.
            Default is True.
        zero_amp_duplicates : bool, optional
            If True, amplitude of duplicated images are forced to be zero.
            Note that it may affect point source ordering!.
            Default is True.
        re_compute : bool, optional
            If True, re-compute (solving the lens equation) image positions,
            even for point source models of type 'IMAGE_POSITIONS'.
            Default is False.

        Returns
        -------
        out : tuple of 2 or 3 1D arrays
            Points in the image plane given as (x-components, y-components),
            along with (optionally) their amplitudes.

        """
        theta_x, theta_y, amps = [], [], []
        for i in self._indices_from_k(k):
            ps = self.point_sources[i]
            ra, dec, amp = ps.image_positions_and_amplitudes(
                kwargs_point_source[i], kwargs_lens=kwargs_lens, kwargs_solver=kwargs_solver,
                zero_amp_duplicates=zero_amp_duplicates, re_compute=re_compute,
            )
            theta_x.append(ra)
            theta_y.append(dec)
            amps.append(amp)

        if with_amplitude:
            return theta_x, theta_y, amps

        return theta_x, theta_y

    def get_source_plane_points(self, kwargs_point_source, kwargs_lens=None,
                                k=None, with_amplitude=True):
        """Compute point source positions and amplitudes in the source plane.

        For point sources defined in the image plane, ray shooting is necessary
        to compute the corresponding source plane positions.

        Parameters
        ----------
        kwargs_point_source : list of dict
            Keyword arguments corresponding to the point source instances.
        kwargs_lens : list of dict, optional
            Keyword arguments for the lensing mass model. Default is None.
        k : int, optional
            Index of the single point source for which to compute positions.
            If None, compute positions for all point sources.
        with_amplitude : bool, optional
            Whether to return the (magnified) amplitude of each point source.
            Default is True.

        Returns
        -------
        out : tuple of 2 or 3 1D arrays
            Points in the source plane given as (x-components, y-components),
            along with (optionally) their amplitudes.

        """
        beta_x, beta_y, amps = [], [], []
        for i in self._indices_from_k(k):
            ps = self.point_sources[i]
            ra, dec = ps.source_position(kwargs_point_source[i], kwargs_lens)
            amp = ps.source_amplitude(kwargs_point_source[i], kwargs_lens)
            beta_x.append(ra)
            beta_y.append(dec)
            amps.append(amp)

        if with_amplitude:
            return beta_x, beta_y, amps

        return beta_x, beta_y
    
    def error_image_plane(self, kwargs_params, kwargs_solver):
        """This function takes as input the current parameters
        and returns distance between the predicted image positions
        based on the mean ray-traced source position, and the model image positions.
        """
        error_image = 0.
        kwargs_point_source = kwargs_params['kwargs_point_source']
        for i, ps_type in enumerate(self.type_list):
            if ps_type == 'IMAGE_POSITIONS':
                ps = self.point_sources[i]
                error_image += ps.error_image_plane(
                    kwargs_params['kwargs_point_source'][i], 
                    kwargs_params['kwargs_lens'], 
                    kwargs_solver,
                )
        return error_image
    
    def log_prob_image_plane(self, kwargs_params, kwargs_solver, **kwargs_hyperparams):
        """This function takes as input the current parameters
        and returns log-probability penalty term that enforces multiply imaged
        point sources optimized following the 'IMAGE_POSITIONS' model
        to be consistent with the lens deflection field.
        """
        log_prob = 0.
        kwargs_point_source = kwargs_params['kwargs_point_source']
        for i, ps_type in enumerate(self.type_list):
            if ps_type == 'IMAGE_POSITIONS':
                ps = self.point_sources[i]
                log_prob += ps.log_prob_image_plane(
                    kwargs_params['kwargs_point_source'][i], 
                    kwargs_params['kwargs_lens'], 
                    kwargs_solver,
                    **kwargs_hyperparams,
                )
        return log_prob
    
    def error_source_plane(self, kwargs_params):
        """This function takes as input the current parameters
        and returns distance between the predicted source positions
        based on the mean source position, and the ray-traced source positions
        from the model image positions.
        """
        error_source = 0.
        kwargs_point_source = kwargs_params['kwargs_point_source']
        for i, ps_type in enumerate(self.type_list):
            if ps_type == 'IMAGE_POSITIONS':
                ps = self.point_sources[i]
                error_source += ps.error_source_plane(
                    kwargs_params['kwargs_point_source'][i], 
                    kwargs_params['kwargs_lens'],
                )
        return error_source
    
    def log_prob_source_plane(self, kwargs_params, **kwargs_hyperparams):
        """This function takes as input the current parameters
        and returns log-probability penalty term that enforces multiply imaged
        point sources optimized following the 'IMAGE_POSITIONS' model
        to be consistent with the lens deflection field.
        """
        log_prob = 0.
        kwargs_point_source = kwargs_params['kwargs_point_source']
        for i, ps_type in enumerate(self.type_list):
            if ps_type == 'IMAGE_POSITIONS':
                ps = self.point_sources[i]
                log_prob += ps.log_prob_source_plane(
                    kwargs_params['kwargs_point_source'][i], 
                    kwargs_params['kwargs_lens'], 
                    **kwargs_hyperparams,
                )
        return log_prob

    def _indices_from_k(self, k):
        """Validate a proposed point source index.

        Parameters
        ----------
        k : int
            Proposed point source index. If k is None or is outside the range
            [0, N - 1], where N is the number of point sources, return the
            list [0, 1, ..., N - 1].

        Returns
        -------
        out : list
            Indices to take from the point source list.

        """
        inds = list(range(len(self.point_sources)))
        if k in inds:
            inds = [k]

        return inds
    


# Utility functions to read simulation products from MOLET (Vernardos et al. 2022)
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal'


import os
import warnings
import numpy as np
from astropy.io import fits

from herculens.Instrument.noise import Noise
from herculens.Instrument.psf import PSF
from herculens.Coordinates.pixel_grid import PixelGrid
from herculens.Util.util import read_json


def read_molet_simulation(molet_path, simu_dir, 
                          use_true_noise_map=False, cut_psf=None,
                          input_file='molet_input.json', 
                          intrument_index=0, instrument_name=None,
                          subtract_offset=True, use_supersampled_psf=False,
                          align_coordinates=True):
    """
    utility method for getting the PixelGrid class from MOLET settings

    If True, align_coordinates will add an offset to the coordinates equal to a supersampled (x10) pixel.
    """
    # load the settings
    input_settings = read_json(os.path.join(molet_path, simu_dir, input_file))
    if instrument_name is None:
        instrument_name = input_settings['instruments'][intrument_index]['name']
        warnings.warn(f"Using MOLET instrument '{instrument_name}'.")
    
    assert input_settings['instruments'][intrument_index]['name'] == instrument_name, "Instrument names are not consistent."
    
    instru_settings = read_json(os.path.join(molet_path, 'instrument_modules', instrument_name, 'specs.json'))
    noise_props = read_json(os.path.join(molet_path, simu_dir, 'output', f'{instrument_name}_noise_properties.json'))
    
    # load data array
    data, data_hdr = fits.getdata(os.path.join(molet_path, simu_dir, 'output', f'OBS_{instrument_name}.fits'), header=True)
    data = data.astype(float)

    # if any offset value was added within MOLET, subtract it back
    offset = 0.
    if 'min_noise' in noise_props:
        if 'offset' not in noise_props:
            warnings.warn("Assuming the constant offset is `abs(min_noise)`!")
        else:
            offset = float(noise_props['offset'])
    
    if subtract_offset is True:
        data -= offset
        if offset != 0.:
            warnings.warn(f"An offset of {offset:.3f} was subtracted from the original MOLET simulation.")

    mass_profiles = input_settings['lenses'][0]['mass_model']
    for mass_profile in mass_profiles:
        if mass_profile['type'] == 'pert':
            pert_file = mass_profile['pars']['filepath']
            pert_path = os.path.join(molet_path, simu_dir, 'input_files', pert_file)
            try:
                dpsi_map = fits.getdata(pert_path, header=False)
            except Exception as e:
                warnings.warn(f"Error when accessing potential perturbation fits file at '{pert_path}':\n{e}")
            else:
                dpsi_map = dpsi_map.astype(float)
        else:
            dpsi_map = None
    
    # load required settings values
    fov_xmin_input = float(input_settings['instruments'][intrument_index]['field-of-view_xmin'])
    # fov_xmax = float(input_settings['instruments'][intrument_index]['field-of-view_xmax'])
    # fov_ymin = float(input_settings['instruments'][intrument_index]['field-of-view_ymin'])
    # fov_ymax = float(input_settings['instruments'][intrument_index]['field-of-view_ymax'])
    fov_xmin = float(data_hdr['XMIN'])
    fov_xmax = float(data_hdr['XMAX'])
    fov_ymin = float(data_hdr['YMIN'])
    fov_ymax = float(data_hdr['YMAX'])
    assert fov_xmin == fov_xmin_input, "Input and output FoV are not consistent."
    
    # the following follows VKL conventions for defining the coordinates grid
    pixel_size = float(instru_settings['resolution'])
    transform_pix2angle = pixel_size * np.eye(2)  # here we assume pixels are square

    width  = fov_xmax - fov_xmin
    height = fov_ymax - fov_ymin
    step_x = pixel_size
    step_y = pixel_size
    #Nx = int(width / step_x + width % step_x)
    #Ny = int(height / step_y + width % step_y)
    Nx = round(width / step_x)
    Ny = round(height / step_y)
    # if Nx % 2 == 1:
    #     Nx += 1
    # if Ny % 2 == 1:
    #     Ny += 1
    ra_at_xy_0 = -width/2. + step_x/2.
    dec_at_xy_0 = -height/2. + step_y/2.
    if align_coordinates:
        half_super_pixel = pixel_size / 10. / 2.  # MOLET uses 10x supersampling
        ra_at_xy_0 += half_super_pixel
        dec_at_xy_0 += half_super_pixel

    # setup the grid class
    kwargs_pixel = {'nx': Nx, 'ny': Ny,
                    'ra_at_xy_0': ra_at_xy_0, 'dec_at_xy_0': dec_at_xy_0, 
                    'transform_pix2angle': transform_pix2angle}
    pixel_grid = PixelGrid(**kwargs_pixel)

    #assert pixel_grid.extent == [fov_xmin + step_x/2., fov_xmax - step_x/2., fov_ymin + step_y/2., fov_ymax - step_y/2.], "Check FoV in MOLET settings."
    assert pixel_grid.pixel_coordinates[0].shape == data.shape, "Shape of image not consistent with coordinates grid."

    # setup the noise class
    noise_type = input_settings['instruments'][intrument_index]['noise']['type']
    if noise_type == 'PoissonNoise':
        background_rms = float(noise_props['sigma_bg'])
        if use_true_noise_map:
            noise_map = fits.getdata(os.path.join(molet_path, simu_dir, 'output', f'{instrument_name}_sigma_map.fits'), header=False)
            noise_map = noise_map.astype(float)
            exp_time  = None
        else:
            noise_map = None
            exp_time  = float(input_settings['instruments'][intrument_index]['noise']['texp'])
    elif noise_type == 'UniformGaussian':
        background_rms = float(noise_props['sigma'])
        exp_time  = None
        noise_map = None
    else:
        raise ValueError(f"Unknown type of noise '{noise_type}'.")
    noise = Noise(Nx, Ny, background_rms=background_rms, exposure_time=exp_time, noise_map=noise_map)

    # if it exists, get the super-sampled PSF that was used for the mock
    super_psf_path = os.path.join(molet_path, simu_dir, 'output', 'supersampled_psf.fits')
    if os.path.exists(super_psf_path):
        psf_kernel_super = fits.getdata(super_psf_path, header=False)
        psf_kernel_super = psf_kernel_super.astype(float)
    else:
        psf_kernel_super = None

    # setup the psf class
    psf_kernel = fits.getdata(os.path.join(molet_path, 'instrument_modules', instrument_name, 'psf.fits'), header=False)
    psf_kernel = psf_kernel.astype(float)
    true_psf_width = float(instru_settings['psf']['width'])
    expe_psf_width = psf_kernel.shape[0] * pixel_size
    if use_supersampled_psf:
        if not len(psf_kernel_super) % 2 == 0:
            psf = PSF(psf_type='PIXEL', kernel_point_source=psf_kernel_super, 
                      point_source_supersampling_factor=10)
        else:
            psf = None
            warnings.warn("Could not prepare the supersampled 'PIXEL' PSF instance as the supersampled PSF is even-sized.")
    else:
        if round(true_psf_width, 7) == round(expe_psf_width, 7):  # means it is not a supersampled PSF
            if cut_psf is not None:
                psf_kernel = psf_kernel[cut_psf:-cut_psf, cut_psf:-cut_psf]
                psf_kernel /= psf_kernel.sum()
            psf = PSF(psf_type='PIXEL', kernel_point_source=psf_kernel)
        else:
            psf = None
            warnings.warn("Could not prepare the 'PIXEL' PSF instance as the PSF in the instrument module is supersampled (what is the supersampling factor?).")
        
    # get specific noise realisation
    noise_real = fits.getdata(os.path.join(molet_path, simu_dir, 'output', f'{instrument_name}_noise_realization.fits'), header=False)
    noise_real = noise_real.astype(float)
    if subtract_offset is True:
        noise_real -= offset

    # load supersampled source and corresponding extent
    source_super, source_hdr = fits.getdata(os.path.join(molet_path, simu_dir, 'output', f'{instrument_name}_source_super.fits'), header=True)
    source_super = source_super.astype(float)
    src_fov_xmin = float(source_hdr['XMIN'])
    src_fov_xmax = float(source_hdr['XMAX'])
    src_fov_ymin = float(source_hdr['YMIN'])
    src_fov_ymax = float(source_hdr['YMAX'])
    # src_extent = (src_fov_xmin, src_fov_xmax, src_fov_ymin, src_fov_ymax)
    src_width  = src_fov_xmax - src_fov_xmin
    src_height = src_fov_ymax - src_fov_ymin
    src_pixel_size = src_width / source_super.shape[0]
    pixel_scale_factor = src_pixel_size / pixel_size
    grid_shape = (src_width, src_height)
    grid_center = ((src_fov_xmin + src_fov_xmax) / 2., (src_fov_ymin + src_fov_ymax) / 2.)
    pixel_grid.create_model_grid(grid_center=grid_center, 
                                 grid_shape=grid_shape, 
                                 pixel_scale_factor=pixel_scale_factor, 
                                 conserve_extent=False,
                                 name='MOLET_source_super')
    #pixel_grid.create_model_grid_simple(source_super.shape, src_extent, name='MOLET_source_super')
    assert pixel_grid.model_pixel_shape('MOLET_source_super') == source_super.shape
    np.testing.assert_almost_equal(pixel_grid.model_pixel_extent('MOLET_source_super'),
                                   [src_fov_xmin + src_pixel_size/2., src_fov_xmax - src_pixel_size/2.,
                                    src_fov_ymin + src_pixel_size/2., src_fov_ymax - src_pixel_size/2.])

    # flux normalisation
    source_super *= pixel_size**2

    return pixel_grid, noise, psf, data, dpsi_map, noise_real, source_super, psf_kernel_super



# Defines a non-singular isothermal ellipsoid
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LensModel.Profiles module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp
import herculens.Util.util as util
import herculens.Util.param_util as param_util


__all__ = ['NIE', 'NIEMajorAxis']


class NIE(object):
    """
    Non-singular isothermal ellipsoid
    kappa = theta_E/2 [s2IE + r2(1 - e * cos(2*phi)]^-1/2
    """
    param_names = ['theta_E', 'e1', 'e2', 's_scale', 'center_x', 'center_y']
    lower_limit_default = {'theta_E': 0, 'e1': -0.5, 'e2': -0.5, 's_scale': 0, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'theta_E': 10, 'e1': 0.5, 'e2': 0.5, 's_scale': 100, 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}
    
    def __init__(self):
        self.nie_major_axis = NIEMajorAxis()
        super(NIE, self).__init__()

    def param_conv(self, theta_E, e1, e2, s_scale):
        """
        convert parameters from 2*kappa = bIE [s2IE + r2(1 - e *cos(2*phi)]^-1/2 to
        2*kappa=  b *(q2(s2 + x2) + y2)^-1/2
        see expressions after Equation 8 in Keeton and Kochanek 1998, https://arxiv.org/pdf/astro-ph/9705194.pdf

        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :return: critical radius b, smoothing scale s, axis ratio q, orientation angle phi_G
        """

        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        theta_E_conv = self._theta_E_q_convert(theta_E, q)
        b = theta_E_conv * jnp.sqrt((1 + q**2)/2)
        s = s_scale * jnp.sqrt((1 + q**2) / (2*q**2))
        return b, s, q, phi_G

    def function(self, x, y, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: lensing potential
        """
        b, s, q, phi_G = self.param_conv(theta_E, e1, e2, s_scale)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f_ = self.nie_major_axis.function(x__, y__, b, s, q)
        # rotate back
        return f_

    def derivatives(self, x, y, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: alpha_x, alpha_y
        """
        b, s, q, phi_G = self.param_conv(theta_E, e1, e2, s_scale)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f__x, f__y = self.nie_major_axis.derivatives(x__, y__, b, s, q)
        # rotate back
        f_x, f_y = util.rotate(f__x, f__y, -phi_G)
        return f_x, f_y

    def hessian(self, x, y, theta_E, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate in image plane
        :param y: y-coordinate in image plane
        :param theta_E: Einstein radius
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale
        :param center_x: profile center
        :param center_y: profile center
        :return: f_xx, f_yy, f_xy
        """
        b, s, q, phi_G = self.param_conv(theta_E, e1, e2, s_scale)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_G)
        # evaluate
        f__xx, f__yy, f__xy = self.nie_major_axis.hessian(x__, y__, b, s, q)
        # rotate back
        kappa = 1./2 * (f__xx + f__yy)
        gamma1__ = 1./2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = jnp.cos(2. * phi_G) * gamma1__ - jnp.sin(2. * phi_G) * gamma2__
        gamma2 = jnp.sin(2. * phi_G) * gamma1__ + jnp.cos(2. * phi_G) * gamma2__
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_yy, f_xy

    def _theta_E_q_convert(self, theta_E, q):
        """
        converts a spherical averaged Einstein radius to an elliptical (major axis) Einstein radius.
        This then follows the convention of the SPEMD profile in lenstronomy.
        (theta_E / theta_E_gravlens) = sqrt[ (1+q^2) / (2 q) ]

        :param theta_E: Einstein radius in lenstronomy conventions
        :param q: axis ratio minor/major
        :return: theta_E in convention of kappa=  b *(q2(s2 + x2) + y2)^-1/2
        """
        theta_E_new = theta_E / (jnp.sqrt((1. + q**2) / (2. * q))) #/ (1+(1-q)/2.)
        return theta_E_new


class NIEMajorAxis(object):
    """
    This class contains the function and the derivatives of the non-singular isothermal ellipse.
    See Keeton and Kochanek 1998, https://arxiv.org/pdf/astro-ph/9705194.pdf

    .. math::
        \kappa =  b *(q2(s2 + x2) + y2)^{-1/2}`

    """

    param_names = ['b', 's', 'q', 'center_x', 'center_y']

    def __init__(self, diff=0.0000000001):  # WARNING: this diff causes large numerical innacuracies in hessian()!
        self._diff = diff
        super(NIEMajorAxis, self).__init__()

    def function(self, x, y, b, s, q):
        psi = self._psi(x, y, q, s)
        alpha_x, alpha_y = self.derivatives(x, y, b, s, q)
        f_ = x * alpha_x + y * alpha_y - b * s * 1. / 2. * jnp.log((psi + s)**2 + (1. - q**2) * x**2)
        return f_

    def derivatives(self, x, y, b, s, q):
        """
        returns df/dx and df/dy of the function
        """
        q = jnp.clip(q, a_max=0.99999999)
        psi = self._psi(x, y, q, s)
        f_x = b / jnp.sqrt(1. - q**2) * jnp.arctan(jnp.sqrt(1. - q**2) * x / (psi + s))
        f_y = b / jnp.sqrt(1. - q**2) * jnp.arctanh(jnp.sqrt(1. - q**2) * y / (psi + q**2 * s))
        return f_x, f_y

    def hessian(self, x, y, b, s, q):
        """
        returns Hessian matrix of function d^2f/dx^2, d^f/dy^2, d^2/dxdy
        """
        alpha_ra, alpha_dec = self.derivatives(x, y, b, s, q)
        diff = self._diff
        alpha_ra_dx, alpha_dec_dx = self.derivatives(x + diff, y, b, s, q)
        alpha_ra_dy, alpha_dec_dy = self.derivatives(x, y + diff, b, s, q)

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        # f_yx = (alpha_dec_dx - alpha_dec)/diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff
        return f_xx, f_yy, f_xy

    @staticmethod
    def kappa(x, y, b, s, q):
        """
        convergence

        :param x: major axis coordinate
        :param y: minor axis coordinate
        :param b: normalization
        :param s: smoothing scale
        :param q: axis ratio
        :return: convergence
        """
        kappa = b/2. * (q**2 * (s**2 + x**2) + y**2)**(-1./2)
        return kappa

    @staticmethod
    def _psi(x, y, q, s):
        """
        expression after equation (8) in Keeton&Kochanek 1998

        :param x: semi-major axis coordinate
        :param y: semi-minor axis coordinate
        :param q: axis ratio minor/major
        :param s: smoothing scale in major axis direction
        :return: phi
        """
        return jnp.sqrt(q**2 * (s**2 + x**2) + y**2)



# Defines a general fully differentiable probability function for inference purposes
# 
# Copyright (c) 2022, herculens developers and contributors

__author__ = 'aymgal'


from functools import partial
import jax.numpy as jnp
from jax import jit, grad, value_and_grad
from tqdm import tqdm


__all__ = ['BaseOptimizer']


class BaseOptimizer(object):

    """Abstract class that defines wraps the loss function for use with optimizers.
    :param loss: herculens.Inference.loss.Loss instance
    :param loss_norm_optim: normalization constant for reducing the magnitude of the loss during optimization.
    """

    def __init__(self, loss, loss_norm_optim=1.):
        self.loss = loss
        self.norm_optim = loss_norm_optim

    def run(self, *args, **kwargs):
        raise NotImplementedError("`run()` method must be implemented.")

    @partial(jit, static_argnums=(0,))
    def function_optim(self, args):
        return self.loss(args) / self.norm_optim

    @partial(jit, static_argnums=(0,))
    def grad_function_optim(self, args):
        return grad(self.loss)(args)

    @partial(jit, static_argnums=(0,))
    def function_optim_with_grad(self, args):
        return value_and_grad(self.function_optim)(args)

    @partial(jit, static_argnums=(0,))
    def function_optim_LM_scalar(self, args):
        return jnp.array([self.function_optim(args)])
    
    @staticmethod
    def _for_loop(iterable, progress_bar_bool, **tqdm_kwargs):
        if progress_bar_bool is True:
            return tqdm(iterable, **tqdm_kwargs)
        else:
            return iterable
 


# Defines a multipole in the potential
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LensModel.Profiles module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'lynevdv', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp

from herculens.MassModel.Profiles.multipole import Multipole as MassMultipole
import herculens.Util.param_util as param_util


__all__ = ['Multipole']


class Multipole(object):
    """
    This class contains a multipole contribution (for 1 component with m>=2)
    This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf
    m : int, multipole order, m>=2
    amp_m : float, multipole strength
    phi_m : float, multipole orientation in radian
    """
    param_names = ['m', 'amp_m', 'phi_m', 'center_x', 'center_y']
    lower_limit_default = {'m': 2,'amp_m':0, 'phi_m': -np.pi, 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'m': 100,'amp_m': 100000, 'phi_m': np.pi, 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}

    def __init__(self):
        self._backend = MassMultipole()

    def function(self, x, y, m, amp_m, phi_m, center_x=0, center_y=0):
        """
        Lensing potential of multipole contribution (for 1 component with m>=2)
        This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf

        :param m: int, multipole order, m>=2
        :param amp_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: x-position
        :return: lensing potential
        """
        return self._backend.function(x, y, m, amp_m, phi_m, center_x=center_x, center_y=center_y)

    def derivatives(self,x,y, m, amp_m, phi_m, center_x=0, center_y=0):
        """
        Deflection of a multipole contribution (for 1 component with m>=2)
        This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf

        :param m: int, multipole order, m>=2
        :param amp_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: x-position
        :return: deflection angles alpha_x, alpha_y
        """
        return self._backend.derivatives(x, y, m, amp_m, phi_m, center_x=center_x, center_y=center_y)



# Handles different method to sample the posterior distribution of parameters
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal', 'austinpeel'


import time
import numpy as np
from functools import partial
import jax

from herculens.Inference.legacy.base_inference import Inference


# TODO: create separate classes for each sampler


__all__ = ['Sampler']


class Sampler(Inference):
    """Class that handles sampling tasks, i.e. approximating posterior distributions of parameters.
    It currently supports:
    - Hamiltonian Monte Carlo using blackjax or numpyro
    - Ensemble Affine Invariant MCMC using emcee
    """

    def hmc_blackjax(self, seed, num_warmup=100, num_samples=100, #num_chains=1, 
                     restart_from_init=False, sampler_type='NUTS', use_stan_warmup=True,
                     step_size=1e-3, inv_mass_matrix=None):
        import blackjax

        rng_key = jax.random.PRNGKey(seed)
        log_prob_fn = self.log_probability
        init_positions = self._param.current_values(as_kwargs=False, restart=restart_from_init)

        if inv_mass_matrix is None:
            # default the inverse mass matrix is the identity matrix
            inv_mass_matrix = np.ones(self._param.num_parameters)

        start = time.time()
        if sampler_type.lower() == 'hmc':
            sampler = blackjax.hmc(log_prob_fn, step_size, inv_mass_matrix, num_samples)
        elif sampler_type.lower() == 'nuts':
            sampler = blackjax.nuts(log_prob_fn, step_size, inv_mass_matrix)
        else:
            raise ValueError(f"Sampler/kernel type '{sampler_type}' is not supported ('NUTS' or 'HMC' only).")

        if use_stan_warmup and sampler_type.lower() == 'nuts':
            rng_key, rng_subkey = jax.random.split(rng_key)
            # here for simplicity we use NUTS
            

            # update step size and inverse mass matrix during warmup with Stan
            window_adaptation = blackjax.window_adaptation(
                blackjax.nuts,
                log_prob_fn, 
                num_steps=num_warmup,
            )
            init_state, kernel, _ = window_adaptation.run(
                rng_key,
                init_positions,
            )
            # reset number of samples so we don't warmup again in the final inference
            num_warmup = 0

        else:
            kernel = jax.jit(sampler.step)
            init_state = sampler.init(init_positions)
        
        # run the inference
        @jax.jit
        def one_step_single_chain(state, rng_key):
            state, info = kernel(rng_key, state)
            return state, (state, info)

        keys = jax.random.split(rng_key, num_warmup + num_samples)
        _, (states, infos) = jax.lax.scan(one_step_single_chain, init_state, keys)
        
        samples = states.position.block_until_ready()
        logL = infos.energy
        runtime = time.time() - start

        if len(samples.shape) == 3:  # basically if num_chains > 1
            # flatten the multiple chains
            s0, s1, s2 = samples.shape
            samples = samples.reshape(s0*s1, s2)
            logL = logL.flatten()

        self._param.set_posterior_samples(samples, logL)
        extra_fields = {
            'step_size': step_size,
            'inverse_mass_matrix': inv_mass_matrix,
            'mean_acceptance_rate': np.mean(infos.acceptance_probability),
            'mean_perc_divergent': 100. * np.mean(infos.is_divergent),
            #'infos': infos,
        }
        return samples, logL, extra_fields, runtime

    def hmc_numpyro(self, seed, num_warmup=100, num_samples=100, num_chains=1, 
                    restart_from_init=False, sampler_type='NUTS', 
                    progress_bar=True, sampler_kwargs={}):
        import numpyro
        #from numpyro.infer.util import ParamInfo

        rng_key = jax.random.PRNGKey(seed)

        if sampler_type.lower() == 'hmc':
            kernel = numpyro.infer.HMC(potential_fn=self.potential_fn, 
                                       kinetic_fn=self.kinetic_fn, 
                                       **sampler_kwargs)
        elif sampler_type.lower() == 'nuts': # NUTS stands for 'no U-turn sampler'
            kernel = numpyro.infer.NUTS(potential_fn=self.potential_fn, 
                                        kinetic_fn=self.kinetic_fn, 
                                        **sampler_kwargs)
        else:
            raise ValueError(f"Sampler/kernel type '{sampler_type}' is not supported ('NUTS' or 'HMC' only).")
        
        init_params = self._param.current_values(as_kwargs=False, restart=restart_from_init)
        # alternative way to provide initial parameters through a NamedTuple:
        #init_params = ParamInfo(init_params, potential_fn(init_params), kinetic_fn(init_params))
        num_dims = len(init_params)
        start = time.time()
        
        # NOTE: disabling the progress-bar can speed up the sampling
        if num_chains > 1:
            init_params = np.repeat(np.expand_dims(init_params, axis=0), num_chains, axis=0)
        mcmc = numpyro.infer.MCMC(kernel, num_warmup=num_warmup, num_samples=num_samples,
                                  num_chains=num_chains, progress_bar=progress_bar)
        mcmc.run(rng_key, init_params=init_params, extra_fields=('potential_energy', 'energy', 'r', 'accept_prob'))
        #mcmc.print_summary(exclude_deterministic=False)
        samples = mcmc.get_samples(group_by_chain=False)
        extra_fields = mcmc.get_extra_fields()

        runtime = time.time() - start
        logL = - extra_fields['potential_energy']
        samples = np.asarray(samples)
        expected_shape = (num_samples*num_chains, num_dims)
        if samples.T.shape == expected_shape:  # this happens sometimes...
            samples = samples.T
            #raise RuntimeError(f"HMC samples do not have correct shape, {samples.shape} instead of {expected_shape}")
        logL = np.asarray(logL)
        self._param.set_posterior_samples(samples, logL)
        return samples, logL, extra_fields, runtime

    def mcmc_emcee(self, log_likelihood_fn, init_stds, walker_ratio=10, 
                   num_warmup=100, num_samples=100, 
                   restart_from_init=False, num_threads=1, progress_bar=True):
        """
        emcee MCMC sampling
        Warning: `log_likelihood_fn` needs to be non.jitted (emcee pickles this function, which is incomptabile with JAX objetcs)
        """
        import emcee

        if num_threads > 1:
            from lenstronomy.Sampling.Pool.pool import choose_pool  # TODO: remove this dependence
            pool = choose_pool(mpi=False, processes=num_threads, use_dill=True)
        else:
            pool = None  # default one
        init_means = self._param.current_values(as_kwargs=False, restart=restart_from_init)
        num_dims = len(init_means)
        num_walkers = int(walker_ratio * num_dims)
        init_params = self._init_emcee_ball(np.asarray(init_means), 
                                            np.asarray(init_stds), 
                                            size=num_walkers, dist='normal')
        start = time.time()
        sampler = emcee.EnsembleSampler(num_walkers, num_dims, log_likelihood_fn,
                                        pool=pool, backend=None)
        sampler.run_mcmc(init_params, num_warmup + num_samples, progress=progress_bar)
        runtime = time.time() - start
        samples = sampler.get_chain(discard=num_warmup, thin=1, flat=True)
        logL = sampler.get_log_prob(flat=True, discard=num_warmup, thin=1)
        extra_fields = None
        self._param.set_posterior_samples(samples, logL)
        return samples, logL, extra_fields, runtime

    @staticmethod
    def _init_emcee_ball(p0, std, size=1, dist='uniform'):
        """
        [from lenstronomy]

        Produce a ball of walkers around an initial parameter value.
        this routine is from the emcee package as it became deprecated there

        :param p0: The initial parameter values (array).
        :param std: The axis-aligned standard deviation (array).
        :param size: The number of samples to produce.
        :param dist: string, specifies the distribution being sampled, supports 'uniform' and 'normal'

        """
        assert(len(p0) == len(std))
        if dist == 'uniform':
            return np.vstack([p0 + std * np.random.uniform(low=-1, high=1, size=len(p0))
                             for i in range(size)])
        elif dist == 'normal':
            return np.vstack([p0 + std * np.random.normal(loc=0, scale=1, size=len(p0))
                              for i in range(size)])
        else:
            raise ValueError('distribution %s not supported. Chose among "uniform" or "normal".' % dist)




# Notes about HMC
# ref: https://bayesianbrad.github.io/posts/2019_hmc.html
# - q is the position, which are variables we are interested in
# - p is the momentum
# The potential energy U(q) will be the minus of the log of the probability density for the distribution 
# of the position variables we wish to sample, plus any constant that is convenient.
# The kinetic energy K(p) will represents the dynamics of our variables.
# A popular form is the Euclidean kinetic energy 1/2 * p^T.(M^-1).p, where M is symmetric, positive definite and typically diagonal.



# Utility functions
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'aymgal'


import copy
import numpy as np
import jax.numpy as jnp
from scipy.ndimage import morphology
from scipy import ndimage
from skimage import measure
from jax import config
from functools import partial

from herculens.LensImage.lens_image import LensImage, LensImage3D
from herculens.LensImage.lens_image_future import LensImage as LensImageFuture
from herculens.LensImage.lens_image_multiplane import MPLensImage
from herculens.LensImage.lensing_operator import LensingOperator


def critical_lines_caustics(lens_image, kwargs_mass, eta_flat=None, supersampling=5, 
                            return_lens_centers=False, k_plane=None):
    if config.read('jax_enable_x64') is not True:
        print("WARNING: JAX's 'jax_enable_x64' is not enabled; "
              "computation of critical lines and caustics might be inaccurate.")
    if isinstance(lens_image, (LensImage, LensImage3D)):
        mass_model = lens_image.MassModel
        inv_mag_fn = partial(mass_model.inverse_magnification, kwargs=kwargs_mass)
        ray_shooting_fn = partial(mass_model.ray_shooting, kwargs=kwargs_mass)
        multiplane_mode = False
    elif isinstance(lens_image, MPLensImage):
        if k_plane is None:
            # we take the furthest plane as the default
            k_plane = lens_image.MPLightModel.number_light_planes
        mass_model = lens_image.MPMassModel
        if eta_flat is None:
            raise ValueError("The `eta_flat` parameter must be provided when using MPLensImage.")
        inv_mag_fn = lambda x, y: mass_model.inverse_magnification(x, y, eta_flat=eta_flat, kwargs=kwargs_mass)[k_plane]
        ray_shooting_fn = lambda x, y: [rs[k_plane] for rs in mass_model.ray_shooting(x, y, eta_flat=eta_flat, kwargs=kwargs_mass)]
        multiplane_mode = True
    else:
        raise TypeError("lens_image must be an instance of LensImage, LensImage3D or MPLensImage.")
    
    # evaluate the inverse magnification
    grid = lens_image.Grid.create_model_grid(pixel_scale_factor=1./supersampling)
    x_grid_img, y_grid_img = grid.pixel_coordinates
    inv_mag_tot = inv_mag_fn(x_grid_img, y_grid_img)
    inv_mag_tot = np.array(inv_mag_tot, dtype=np.float64)

    # find contours corresponding to infinite magnification
    contours = measure.find_contours(inv_mag_tot, 0.)

    crit_lines, caustics = [], []
    for contour in contours:
        # extract the lines
        cline_x, cline_y = contour[:, 1], contour[:, 0]
        # convert to model coordinates
        cline_x, cline_y = grid.map_pix2coord(cline_x, cline_y)
        crit_lines.append((np.array(cline_x), np.array(cline_y)))
        # find corresponding caustics through ray shooting
        caust_x, caust_y = ray_shooting_fn(cline_x, cline_y)
        caustics.append((np.array(caust_x), np.array(caust_y)))

    # can also returns the lens components centroids for convenience
    if return_lens_centers:
        cxs, cys = [], []
        if multiplane_mode is True:
            k_main_lens = 0
            for kw in kwargs_mass[k_main_lens]:
                if 'center_x' in kw:
                    cxs.append(kw['center_x'])
                    cys.append(kw['center_y'])
        else:
            for kw in kwargs_mass:
                if 'center_x' in kw:
                    cxs.append(kw['center_x'])
                    cys.append(kw['center_y'])
        return crit_lines, caustics, (np.array(cxs), np.array(cys))
    return crit_lines, caustics


def shear_deflection_field(lens_image, kwargs_lens, num_pixels=20):
    from herculens.MassModel.Profiles.shear import Shear, ShearGammaPsi
    shear_idx, shear_type = None, None
    num_profiles = len(lens_image.MassModel.profile_type_list)
    assert num_profiles == len(kwargs_lens)
    for i in range(num_profiles):
        if (lens_image.MassModel.profile_type_list[i] == 'SHEAR' or
            isinstance(lens_image.MassModel.profile_type_list[i], Shear)):
            shear_idx = i
            shear_type = 'SHEAR'
            break
        elif (lens_image.MassModel.profile_type_list[i] == 'SHEAR_GAMMA_PSI' or
              isinstance(lens_image.MassModel.profile_type_list[i], ShearGammaPsi)):
            shear_idx = i
            shear_type = 'SHEAR_GAMMA_PSI'
            break
    if shear_idx is None:
        return None
    if shear_type == 'SHEAR_GAMMA_PSI':
        phi_ext, gamma_ext = kwargs_lens[shear_idx]['phi_ext'], kwargs_lens[shear_idx]['gamma_ext']
    else:
        # imports are here to avoid issues with circular imports
        from herculens.Util.param_util import shear_cartesian2polar_numpy
        phi_ext, gamma_ext = shear_cartesian2polar_numpy(
            kwargs_lens[shear_idx]['gamma1'], 
            kwargs_lens[shear_idx]['gamma2'],
        )
    grid = lens_image.Grid.create_model_grid(num_pixels=num_pixels)
    x_grid_img, y_grid_img = grid.pixel_coordinates
    gamma_x = gamma_ext*np.cos(phi_ext)
    gamma_y = gamma_ext*np.sin(phi_ext)
    if gamma_x.size == 1:
        gamma_x = np.full_like(x_grid_img, float(gamma_x))
        gamma_y = np.full_like(y_grid_img, float(gamma_y))
    return (np.array(x_grid_img), np.array(y_grid_img), gamma_x, gamma_y)


def total_shear_deflection_field(lens_image, kwargs_mass, eta_flat=None, num_pixels=20, k_plane=-1):
    if isinstance(lens_image, (LensImage, LensImage3D)):
        mass_model = lens_image.MassModel
        gamma_fn = partial(mass_model.gamma, kwargs_lens=kwargs_mass)
        multiplane_mode = False
    elif isinstance(lens_image, MPLensImage):
        mass_model = lens_image.MPMassModel
        if eta_flat is None:
            raise ValueError("The `eta_flat` parameter must be provided when using MPLensImage.")
        gamma_fn = partial(mass_model.gamma, eta_flat=eta_flat, kwargs=kwargs_mass)
        multiplane_mode = True
    else:
        raise TypeError("lens_image must be an instance of LensImage, LensImage3D or MPLensImage.")
    grid = lens_image.Grid.create_model_grid(num_pixels=num_pixels)
    x_grid_img, y_grid_img = grid.pixel_coordinates
    gamma_x, gamma_y = gamma_fn(x_grid_img, y_grid_img)
    if multiplane_mode is True:
        gamma_x = gamma_x[k_plane]
        gamma_y = gamma_y[k_plane]
    if gamma_x.size == 1:
        gamma_x = np.full_like(x_grid_img, float(gamma_x))
        gamma_y = np.full_like(y_grid_img, float(gamma_y))
    return (np.array(x_grid_img), np.array(y_grid_img), gamma_x, gamma_y)


def _get_parameters(parameters):
    # For backward compatibility, we need to handle both the old (legacy) Parameters class
    # and the simpler way to provide parameters (which is a simple dictionary)
    if isinstance(parameters, dict):
        kwargs_params = parameters
    else:
        from herculens.Inference.legacy.parameters import Parameters as LegacyParameters
        if not isinstance(parameters, LegacyParameters):
            raise TypeError("The 'parameters' argument must be either a dictionary or a Parameters instance.")
        kwargs_params = copy.deepcopy(parameters.current_values(as_kwargs=True))
    return kwargs_params


def mask_from_source_area(lens_image, parameters):
    src_idx = lens_image.SourceModel.pixelated_index
    kwargs_param_mask = _get_parameters(parameters)
    pixels = kwargs_param_mask['kwargs_source'][src_idx]['pixels']
    pixels = np.zeros_like(pixels)
    pixels[3:-3, 3:-3] = 1.  # so the source plane is ones except in a small margin all around
    kwargs_param_mask['kwargs_source'][src_idx]['pixels'] = jnp.array(pixels)
    model_mask = lens_image.source_surface_brightness(kwargs_param_mask['kwargs_source'],
                                                      kwargs_lens=kwargs_param_mask['kwargs_lens'],
                                                      unconvolved=True, de_lensed=False)
    model_mask = np.array(model_mask)
    model_mask[model_mask < 0.1] = 0.
    model_mask[model_mask >= 0.1] = 1.
    model_mask = morphology.binary_opening(model_mask, iterations=10)
    model_mask = morphology.binary_dilation(model_mask, iterations=3)
    return model_mask


def mask_from_lensed_source(lens_image, parameters=None, source_model=None,
                            threshold=0.1, smoothing=0, kwargs_numerics=None):
    # imports are here to avoid issues with circular imports
    from herculens.LensImage.lens_image import LensImage
    from herculens.LightModel.light_model import LightModel

    if parameters is None and source_model is None:
        raise ValueError("You must provide a source model "
                         "if no parameters are provided.")
    if parameters is not None:
        kwargs_param = _get_parameters(parameters)
        source_model = lens_image.source_surface_brightness(kwargs_param['kwargs_source'], 
                                                            de_lensed=True, unconvolved=True)
    source_model = np.array(source_model)
    if smoothing > 0:
        source_model = ndimage.gaussian_filter(source_model, sigma=smoothing)
    binary_source = source_model / source_model.max()
    binary_source[binary_source < threshold] = 0.
    binary_source[binary_source >= threshold] = 1.
    grid = copy.deepcopy(lens_image.Grid)
    lens_image_pixel = LensImage(grid, lens_image.PSF, 
                                 noise_class=lens_image.Noise,
                                 lens_mass_model_class=lens_image.MassModel,
                                 source_model_class=LightModel(['PIXELATED']),
                                 lens_light_model_class=lens_image.LensLightModel,
                                 kwargs_numerics=kwargs_numerics)
    kwargs_param_mask = copy.deepcopy(kwargs_param)
    kwargs_param_mask['kwargs_source'] = [{'pixels': jnp.array(binary_source)}]
    model_mask = lens_image_pixel.source_surface_brightness(kwargs_param_mask['kwargs_source'], 
                                                      kwargs_lens=kwargs_param_mask['kwargs_lens'],
                                                      unconvolved=True, de_lensed=False)
    model_mask = np.array(model_mask)
    model_mask[model_mask < 0.1] = 0.
    model_mask[model_mask >= 0.1] = 1.
    #model_mask = morphology.binary_opening(model_mask, iterations=10)
    #model_mask = morphology.binary_dilation(model_mask, iterations=3)
    return model_mask, binary_source


def pixelated_region_from_sersic(kwargs_sersic, force_square=False, use_major_axis=False,
                                 min_width=1.0, min_height=1.0, scaling=1.0):
    # imports are here to avoid issues with circular imports
    from herculens.Util import param_util

    # TODO: support arbitrary smooth source profile
    c_x = kwargs_sersic['center_x']
    c_y = kwargs_sersic['center_y']
    r_eff = kwargs_sersic['R_sersic']
    if 'e1' in kwargs_sersic:
        e1 = kwargs_sersic['e1']
        e2 = kwargs_sersic['e2']
    else:
        e1 = e2 = 0.
    phi, q = param_util.ellipticity2phi_q(e1, e2)
    a = r_eff / np.sqrt(q)  # semi-major axis
    r = r_eff * np.sqrt(q)  # product average radius
    print(f"r_eff={r_eff:.2f}, r={r:.2f}, a={a:.2f}")
    diameter = 2*a if use_major_axis else 2*r
    width = diameter * np.abs(np.cos(phi)) * scaling
    height = diameter * np.abs(np.sin(phi)) * scaling
    width = max(min_width, width)
    height = max(min_height, height)
    if force_square is True:
        width = max(width, height)
        height = width
    # the following dict is consistent with arguments of PixelGrid.create_model_grid()
    kwargs_pixelated_grid = {
        'grid_center': (float(c_x), float(c_y)),
        'grid_shape': (float(width), float(height)),
    }
    return kwargs_pixelated_grid


def pixelated_region_from_arc_mask(arc_mask, image_grid, mass_model, mass_params):
    # We first design a hypothetical source plane with sufficiently high resolution
    nx, ny = image_grid.num_pixel_axes
    nx_src, ny_src = (nx//3, ny//3)
    high_res_source_grid = image_grid.create_model_grid(pixel_scale_factor=0.5, grid_shape=(nx_src, ny_src))
    
    # Then build the lensing operator based on the image/source grids and mass model
    lensing_op = LensingOperator(mass_model, image_grid, high_res_source_grid)
    lensing_op.compute_mapping(kwargs_lens=mass_params)
    
    # De-lens the arc mask
    arc_mask_source = np.array(lensing_op.image2source_2d(arc_mask))
    arc_mask_source = np.where(arc_mask_source < 0.1, 0., 1.)  # only contains 0 and 1 now

    # Find the minimum bounding box that encloses the mask in source plane (square here)
    # - get the extrema 
    rows = np.max(arc_mask_source, axis=0)
    print(rows)
    row_low, row_high = np.argmax(rows), np.argmax(rows[::-1])
    print("R", row_low, row_high)
    cols = np.max(arc_mask_source, axis=1)
    print(cols)
    col_low, col_high = np.argmax(cols), np.argmax(cols[::-1])
    print("C", col_low, col_high)
    # - get the pixel coordinates corresponding to these extrema
    x_low, y_low = high_res_source_grid.map_pix2coord(row_low, col_low)
    x_high, y_high = high_res_source_grid.map_pix2coord(row_high, col_high)
    print(x_low, y_low)
    print(x_high, y_high)

    # Get the grid parameters
    new_grid_shape = (abs(x_high - x_low), abs(y_high - y_low))
    new_grid_center = (0.5*(x_low+x_high), 0.5*(y_low+y_high))
    kwargs_pixelated_grid = {
        'grid_center': new_grid_center,
        'grid_shape': new_grid_shape,
    }

    # Create the new, reduced source plane grid
    return kwargs_pixelated_grid


def estimate_model_covariance(lens_image, parameters, samples, return_cross_covariance=False):
    model_samples = []
    for sample in samples:
        kwargs_sample = parameters.args2kwargs(sample)
        model_sample = lens_image.model(**kwargs_sample)
        model_map_shape = model_sample.shape
        model_samples.append(model_sample.flatten())
    model_samples = np.array(model_samples)
    
    # variance map (as a 2D image)
    model_var_map = np.var(model_samples, axis=0).reshape(*model_map_shape)
    
    # full (auto) covariance matrix from the model samples
    model_cov = np.cov(model_samples, rowvar=False)

    # computation of the cross-covariance matrix from model and data samples
    if return_cross_covariance:
        data_mean_proxy = lens_image.model(**parameters.best_fit_values(as_kwargs=True))   # best-fit model as a proxy for the unknown data 'mean'
        data_var = lens_image.Noise.C_D_model(data_mean_proxy, force_recompute=True)
        data_mean_proxy = data_mean_proxy.flatten()
        data_cov = np.diag(data_var.flatten())
        data_samples = draw_samples_from_covariance(data_mean_proxy, data_cov, num_samples=len(model_samples))
        data_vector  = np.mean(data_samples - data_samples.mean(axis=0), axis=0)
        model_vector = np.mean(model_samples - model_samples.mean(axis=0), axis=0)
        data_model_cross_cov = np.outer(data_vector, model_vector)
        return model_var_map, model_cov, data_model_cross_cov
    else:
        return model_var_map, model_cov


def draw_samples_from_covariance(mean, covariance, num_samples=10000, seed=None):
    if seed is not None:
        np.random.seed(seed)
    samples = np.random.multivariate_normal(mean, covariance, size=num_samples)
    return samples

